import collections
import io
import json
import logging
import math
import numbers
import os
import time
import warnings
from asyncio import iscoroutinefunction
from collections import OrderedDict, abc
from contextlib import contextmanager
from copy import deepcopy
from functools import partial
from inspect import signature
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Hashable,
    Iterator,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Sequence,
    SupportsAbs,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

import numpy as np

if TYPE_CHECKING:
    from PyQt5.QtWidgets import QMainWindow

QCODES_USER_PATH_ENV = 'QCODES_USER_PATH'


_tprint_times: Dict[str, float] = {}


log = logging.getLogger(__name__)


class NumpyJSONEncoder(json.JSONEncoder):
    """
    This JSON encoder adds support for serializing types that the built-in
    ``json`` module does not support out-of-the-box. See the docstring of the
    ``default`` method for the description of all conversions.
    """

    def default(self, obj: Any) -> Any:
        """
        List of conversions that this encoder performs:

        * ``numpy.generic`` (all integer, floating, and other types) gets
          converted to its python equivalent using its ``item`` method (see
          ``numpy`` docs for more information,
          https://docs.scipy.org/doc/numpy/reference/arrays.scalars.html).
        * ``numpy.ndarray`` gets converted to python list using its ``tolist``
          method.
        * Complex number (a number that conforms to ``numbers.Complex`` ABC) gets
          converted to a dictionary with fields ``re`` and ``im`` containing floating
          numbers for the real and imaginary parts respectively, and a field
          ``__dtype__`` containing value ``complex``.
        * Numbers with uncertainties  (numbers that conforms to ``uncertainties.UFloat``) get
          converted to a dictionary with fields ``nominal_value`` and ``std_dev`` containing floating
          numbers for the nominal and uncertainty parts respectively, and a field
          ``__dtype__`` containing value ``UFloat``.
        * Object with a ``_JSONEncoder`` method get converted the return value of
          that method.
        * Objects which support the pickle protocol get converted using the
          data provided by that protocol.
        * Other objects which cannot be serialized get converted to their
          string representation (using the ``str`` function).
        """
        with warnings.catch_warnings():
            # this context manager can be removed when uncertainties
            # no longer triggers deprecation warnings
            warnings.simplefilter("ignore", category=DeprecationWarning)
            import uncertainties

        if isinstance(obj, np.generic) \
                and not isinstance(obj, np.complexfloating):
            # for numpy scalars
            return obj.item()
        elif isinstance(obj, np.ndarray):
            # for numpy arrays
            return obj.tolist()
        elif (isinstance(obj, numbers.Complex) and
              not isinstance(obj, numbers.Real)):
            return {
                '__dtype__': 'complex',
                're': float(obj.real),
                'im': float(obj.imag)
            }
        elif isinstance(obj, uncertainties.UFloat):
            return {
                '__dtype__': 'UFloat',
                'nominal_value': float(obj.nominal_value),
                'std_dev': float(obj.std_dev)
            }
        elif hasattr(obj, '_JSONEncoder'):
            # Use object's custom JSON encoder
            jsosencode = getattr(obj, "_JSONEncoder")
            return jsosencode()
        else:
            try:
                s = super().default(obj)
            except TypeError:
                # json does not support dumping UserDict but
                # we can dump the dict stored internally in the
                # UserDict
                if isinstance(obj, collections.UserDict):
                    return obj.data
                # See if the object supports the pickle protocol.
                # If so, we should be able to use that to serialize.
                if hasattr(obj, '__getnewargs__'):
                    return {
                        '__class__': type(obj).__name__,
                        '__args__': getattr(obj, "__getnewargs__")()
                    }
                else:
                    # we cannot convert the object to JSON, just take a string
                    s = str(obj)
            return s


def tprint(string: str, dt: int = 1, tag: str = 'default') -> None:
    """Print progress of a loop every ``dt`` seconds."""
    ptime = _tprint_times.get(tag, 0)
    if (time.time() - ptime) > dt:
        print(string)
        _tprint_times[tag] = time.time()


def is_sequence(obj: Any) -> bool:
    """
    Test if an object is a sequence.

    We do not consider strings or unordered collections like sets to be
    sequences, but we do accept iterators (such as generators).
    """
    return (isinstance(obj, (abc.Iterator, abc.Sequence, np.ndarray)) and
            not isinstance(obj, (str, bytes, io.IOBase)))


def is_sequence_of(obj: Any,
                   types: Optional[Union[Type[object],
                                         Tuple[Type[object], ...]]] = None,
                   depth: Optional[int] = None,
                   shape: Optional[Sequence[int]] = None
                   ) -> bool:
    """
    Test if object is a sequence of entirely certain class(es).

    Args:
        obj: The object to test.
        types: Allowed type(s). If omitted, we just test the depth/shape.
        depth: Level of nesting, ie if ``depth=2`` we expect a sequence of
               sequences. Default 1 unless ``shape`` is supplied.
        shape: The shape of the sequence, ie its length in each dimension.
               If ``depth`` is omitted, but ``shape`` included, we set
               ``depth = len(shape)``.

    Returns:
        bool: ``True`` if every item in ``obj`` matches ``types``.
    """
    if not is_sequence(obj):
        return False

    if shape is None or shape == ():
        next_shape: Optional[Tuple[int]] = None
        if depth is None:
            depth = 1
    else:
        if depth is None:
            depth = len(shape)
        elif depth != len(shape):
            raise ValueError('inconsistent depth and shape')

        if len(obj) != shape[0]:
            return False

        next_shape = cast(Tuple[int], shape[1:])

    for item in obj:
        if depth > 1:
            if not is_sequence_of(item, types, depth=depth - 1,
                                  shape=next_shape):
                return False
        elif types is not None and not isinstance(item, types):
            return False
    return True


def is_function(f: Callable[..., Any],
                arg_count: int, coroutine: bool = False) -> bool:
    """
    Check and require a function that can accept the specified number of
    positional arguments, which either is or is not a coroutine
    type casting "functions" are allowed, but only in the 1-argument form.

    Args:
        f: Function to check.
        arg_count: Number of argument f should accept.
        coroutine: Is a coroutine.

    Return:
        bool: is function and accepts the specified number of arguments.

    """
    if not isinstance(arg_count, int) or arg_count < 0:
        raise TypeError('arg_count must be a non-negative integer')

    if not (callable(f) and bool(coroutine) is iscoroutinefunction(f)):
        return False

    if isinstance(f, type):
        # for type casting functions, eg int, str, float
        # only support the one-parameter form of these,
        # otherwise the user should make an explicit function.
        return arg_count == 1

    try:
        sig = signature(f)
    except ValueError:
        # some built-in functions/methods don't describe themselves to inspect
        # we already know it's a callable and coroutine is correct.
        return True

    try:
        inputs = [0] * arg_count
        sig.bind(*inputs)
        return True
    except TypeError:
        return False


def full_class(obj: object) -> str:
    """The full importable path to an object's class."""
    return type(obj).__module__ + '.' + type(obj).__name__


def named_repr(obj: Any) -> str:
    """Enhance the standard repr() with the object's name attribute."""
    s = '<{}.{}: {} at {}>'.format(
        obj.__module__,
        type(obj).__name__,
        str(obj.name),
        id(obj))
    return s


K = TypeVar('K', bound=Hashable)
L = TypeVar('L', bound=Hashable)


def deep_update(
        dest: MutableMapping[K, Any],
        update: Mapping[L, Any]
) -> MutableMapping[Union[K, L], Any]:
    """
    Recursively update one JSON structure with another.

    Only dives into nested dicts; lists get replaced completely.
    If the original value is a dictionary and the new value is not, or vice versa,
    we also replace the value completely.
    """
    dest_int = cast(MutableMapping[Union[K, L], Any], dest)
    for k, v_update in update.items():
        v_dest = dest_int.get(k)
        if isinstance(v_update, abc.Mapping) and isinstance(v_dest, abc.MutableMapping):
            deep_update(v_dest, v_update)
        else:
            dest_int[k] = deepcopy(v_update)
    return dest_int


# could use numpy.arange here, but
# a) we don't want to require that as a dep so low level
# b) I'd like to be more flexible with the sign of step
def permissive_range(start: float, stop: float, step: SupportsAbs[float]
                     ) -> List[float]:
    """
    Returns a range (as a list of values) with floating point steps.
    Always starts at start and moves toward stop, regardless of the
    sign of step.

    Args:
        start: The starting value of the range.
        stop: The end value of the range.
        step: Spacing between the values.
    """
    signed_step = abs(step) * (1 if stop > start else -1)
    # take off a tiny bit for rounding errors
    step_count = math.ceil((stop - start) / signed_step - 1e-10)
    return [start + i * signed_step for i in range(step_count)]


# This is very much related to the permissive_range but more
# strict on the input, start and endpoints are always included,
# and a sweep is only created if the step matches an integer
# number of points.
# numpy is a dependency anyways.
# Furthermore the sweep allows to take a number of points and generates
# an array with endpoints included, which is more intuitive to use in a sweep.
def make_sweep(start: float,
               stop: float,
               step: Optional[float] = None,
               num: Optional[int] = None
               ) -> List[float]:
    """
    Generate numbers over a specified interval.
    Requires ``start`` and ``stop`` and (``step`` or ``num``).
    The sign of ``step`` is not relevant.

    Args:
        start: The starting value of the sequence.
        stop: The end value of the sequence.
        step:  Spacing between values.
        num: Number of values to generate.

    Returns:
        numpy.ndarray: numbers over a specified interval as a ``numpy.linspace``.

    Examples:
        >>> make_sweep(0, 10, num=5)
        [0.0, 2.5, 5.0, 7.5, 10.0]
        >>> make_sweep(5, 10, step=1)
        [5.0, 6.0, 7.0, 8.0, 9.0, 10.0]
        >>> make_sweep(15, 10.5, step=1.5)
        >[15.0, 13.5, 12.0, 10.5]
    """
    if step and num:
        raise AttributeError('Don\'t use `step` and `num` at the same time.')
    if (step is None) and (num is None):
        raise ValueError('If you really want to go from `start` to '
                         '`stop` in one step, specify `num=2`.')
    if step is not None:
        steps = abs((stop - start) / step)
        tolerance = 1e-10
        steps_lo = int(np.floor(steps + tolerance))
        steps_hi = int(np.ceil(steps - tolerance))

        if steps_lo != steps_hi:
            raise ValueError(
                'Could not find an integer number of points for '
                'the the given `start`, `stop`, and `step` '
                'values. \nNumber of points is {:d} or {:d}.'
                .format(steps_lo + 1, steps_hi + 1))
        num_steps = steps_lo + 1
    elif num is not None:
        num_steps = num

    output_list = np.linspace(start, stop, num=num_steps).tolist()
    return cast(List[float], output_list)


def wait_secs(finish_clock: float) -> float:
    """
    Calculate the number of seconds until a given clock time.
    The clock time should be the result of ``time.perf_counter()``.
    Does NOT wait for this time.
    """
    delay = finish_clock - time.perf_counter()
    if delay < 0:
        logging.warning(f'negative delay {delay:.6f} sec')
        return 0
    return delay


class DelegateAttributes:
    """
    Mixin class to create attributes of this object by
    delegating them to one or more dictionaries and/or objects.

    Also fixes ``__dir__`` so the delegated attributes will show up
    in ``dir()`` and ``autocomplete``.

    Attribute resolution order:
        1. Real attributes of this object.
        2. Keys of each dictionary in ``delegate_attr_dicts`` (in order).
        3. Attributes of each object in ``delegate_attr_objects`` (in order).
    """
    delegate_attr_dicts: List[str] = []
    """
    A list of names (strings) of dictionaries
    which are (or will be) attributes of ``self``, whose keys should
    be treated as attributes of ``self``.
    """
    delegate_attr_objects: List[str] = []
    """
    A list of names (strings) of objects
    which are (or will be) attributes of ``self``, whose attributes
    should be passed through to ``self``.
    """
    omit_delegate_attrs: List[str] = []
    """
    A list of attribute names (strings)
    to *not* delegate to any other dictionary or object.
    """

    def __getattr__(self, key: str) -> Any:
        if key in self.omit_delegate_attrs:
            raise AttributeError("'{}' does not delegate attribute {}".format(
                self.__class__.__name__, key))

        for name in self.delegate_attr_dicts:
            if key == name:
                # needed to prevent infinite loops!
                raise AttributeError(
                    "dict '{}' has not been created in object '{}'".format(
                        key, self.__class__.__name__))
            try:
                d = getattr(self, name, None)
                if d is not None:
                    return d[key]
            except KeyError:
                pass

        for name in self.delegate_attr_objects:
            if key == name:
                raise AttributeError(
                    "object '{}' has not been created in object '{}'".format(
                        key, self.__class__.__name__))
            try:
                obj = getattr(self, name, None)
                if obj is not None:
                    return getattr(obj, key)
            except AttributeError:
                pass

        raise AttributeError(
            "'{}' object and its delegates have no attribute '{}'".format(
                self.__class__.__name__, key))

    def __dir__(self) -> List[str]:
        names = list(super().__dir__())
        for name in self.delegate_attr_dicts:
            d = getattr(self, name, None)
            if d is not None:
                names += [k for k in d.keys()
                          if k not in self.omit_delegate_attrs]

        for name in self.delegate_attr_objects:
            obj = getattr(self, name, None)
            if obj is not None:
                names += [k for k in dir(obj)
                          if k not in self.omit_delegate_attrs]

        return sorted(set(names))


def strip_attrs(obj: object, whitelist: Sequence[str] = ()) -> None:
    """
    Irreversibly remove all direct instance attributes of object, to help with
    disposal, breaking circular references.

    Args:
        obj: Object to be stripped.
        whitelist: List of names that are not stripped from the object.
    """
    try:
        lst = set(list(obj.__dict__.keys())) - set(whitelist)
        for key in lst:
            try:
                del obj.__dict__[key]
            # TODO (giulioungaretti) fix bare-except
            except:
                pass
        # TODO (giulioungaretti) fix bare-except
    except:
        pass


def compare_dictionaries(dict_1: Dict[Hashable, Any],
                         dict_2: Dict[Hashable, Any],
                         dict_1_name: Optional[str] = 'd1',
                         dict_2_name: Optional[str] = 'd2',
                         path: str = "") -> Tuple[bool, str]:
    """
    Compare two dictionaries recursively to find non matching elements.

    Args:
        dict_1: First dictionary to compare.
        dict_2: Second dictionary to compare.
        dict_1_name: Optional name of the first dictionary used in the
                     differences string.
        dict_2_name: Optional name of the second dictionary used in the
                     differences string.
    Returns:
        Tuple: Are the dicts equal and the difference rendered as
               a string.

    """
    err = ''
    key_err = ''
    value_err = ''
    old_path = path
    for k in dict_1.keys():
        path = old_path + "[%s]" % k
        if k not in dict_2.keys():
            key_err += "Key {}{} not in {}\n".format(
                dict_1_name, path, dict_2_name)
        else:
            if isinstance(dict_1[k], dict) and isinstance(dict_2[k], dict):
                err += compare_dictionaries(dict_1[k], dict_2[k],
                                            dict_1_name, dict_2_name, path)[1]
            else:
                match = (dict_1[k] == dict_2[k])

                # if values are equal-length numpy arrays, the result of
                # "==" is a bool array, so we need to 'all' it.
                # In any other case "==" returns a bool
                # TODO(alexcjohnson): actually, if *one* is a numpy array
                # and the other is another sequence with the same entries,
                # this will compare them as equal. Do we want this, or should
                # we require exact type match?
                if hasattr(match, 'all'):
                    match = match.all()

                if not match:
                    value_err += (
                        'Value of "{}{}" ("{}", type"{}") not same as\n'
                        '  "{}{}" ("{}", type"{}")\n\n').format(
                        dict_1_name, path, dict_1[k], type(dict_1[k]),
                        dict_2_name, path, dict_2[k], type(dict_2[k]))

    for k in dict_2.keys():
        path = old_path + f"[{k}]"
        if k not in dict_1.keys():
            key_err += "Key {}{} not in {}\n".format(
                dict_2_name, path, dict_1_name)

    dict_differences = key_err + value_err + err
    if len(dict_differences) == 0:
        dicts_equal = True
    else:
        dicts_equal = False
    return dicts_equal, dict_differences


def warn_units(class_name: str, instance: object) -> None:
    logging.warning('`units` is deprecated for the `' + class_name +
                    '` class, use `unit` instead. ' + repr(instance))


def foreground_qt_window(window: "QMainWindow") -> None:
    """
    Try as hard as possible to bring a qt window to the front. This
    will use pywin32 if installed and running on windows as this
    seems to be the only reliable way to foreground a window. The
    build-in qt functions often doesn't work. Note that to use this
    with pyqtgraphs remote process you should use the ref in that module
    as in the example below.

    Args:
        window: Handle to qt window to foreground.
    Examples:
        >>> Qtplot.qt_helpers.foreground_qt_window(plot.win)
    """
    try:
        import win32con
        from win32gui import SetWindowPos

        # use the idea from
        # https://stackoverflow.com/questions/12118939/how-to-make-a-pyqt4-window-jump-to-the-front
        SetWindowPos(window.winId(),
                     win32con.HWND_TOPMOST, # = always on top. only reliable way to bring it to the front on windows
                     0, 0, 0, 0,
                     win32con.SWP_NOMOVE | win32con.SWP_NOSIZE | win32con.SWP_SHOWWINDOW)
        SetWindowPos(window.winId(),
                     win32con.HWND_NOTOPMOST, # disable the always on top, but leave window at its top position
                     0, 0, 0, 0,
                     win32con.SWP_NOMOVE | win32con.SWP_NOSIZE | win32con.SWP_SHOWWINDOW)
    except ImportError:
        pass
    window.show()
    window.raise_()
    window.activateWindow()


def add_to_spyder_UMR_excludelist(modulename: str) -> None:
    """
    Spyder tries to reload any user module. This does not work well for
    qcodes because it overwrites Class variables. QCoDeS uses these to
    store global attributes such as default station, monitor and list of
    instruments. This "feature" can be disabled by the
    gui. Unfortunately this cannot be disabled in a natural way
    programmatically so in this hack we replace the global ``__umr__`` instance
    with a new one containing the module we want to exclude. This will do
    nothing if Spyder is not found.
    TODO is there a better way to detect if we are in spyder?
    """
    if any('SPYDER' in name for name in os.environ):

        sitecustomize_found = False
        try:
            from spyder.utils.site import sitecustomize
        except ImportError:
            pass
        else:
            sitecustomize_found = True
        if sitecustomize_found is False:
            try:
                from spyder_kernels.customize import spydercustomize as sitecustomize

            except ImportError:
                pass
            else:
                sitecustomize_found = True

        if sitecustomize_found is False:
            return

        excludednamelist = os.environ.get('SPY_UMR_NAMELIST',
                                          '').split(',')
        if modulename not in excludednamelist:
            log.info(f"adding {modulename} to excluded modules")
            excludednamelist.append(modulename)
            sitecustomize.__umr__ = sitecustomize.UserModuleReloader(namelist=excludednamelist)
            os.environ['SPY_UMR_NAMELIST'] = ','.join(excludednamelist)


@contextmanager
def attribute_set_to(object_: object,
                     attribute_name: str,
                     new_value: Any) -> Iterator[None]:
    """
    This context manager allows to change a given attribute of a given object
    to a new value, and the original value is reverted upon exit of the context
    manager.

    Args:
        object_: The object which attribute value is to be changed.
        attribute_name: The name of the attribute that is to be changed.
        new_value: The new value to which the attribute of the object is
                   to be changed.
    """
    old_value = getattr(object_, attribute_name)
    setattr(object_, attribute_name, new_value)
    try:
        yield
    finally:
        setattr(object_, attribute_name, old_value)


def partial_with_docstring(func: Callable[..., Any],
                           docstring: str,
                           **kwargs: Any) -> Callable[..., Any]:
    """
    We want to have a partial function which will allow us access the docstring
    through the python built-in help function. This is particularly important
    for client-facing driver methods, whose arguments might not be obvious.

    Consider the follow example why this is needed:

    >>> from functools import partial
    >>> def f():
    >>> ... pass
    >>> g = partial(f)
    >>> g.__doc__ = "bla"
    >>> help(g) # this will print an unhelpful message

    Args:
        func: A function that its docstring will be accessed.
        docstring: The docstring of the corresponding function.
    """
    ex = partial(func, **kwargs)

    def inner(**inner_kwargs: Any) -> None:
        ex(**inner_kwargs)

    inner.__doc__ = docstring

    return inner


def create_on_off_val_mapping(on_val: Any = True, off_val: Any = False
                              ) -> Dict[Union[str, bool], Any]:
    """
    Returns a value mapping which maps inputs which reasonably mean "on"/"off"
    to the specified ``on_val``/``off_val`` which are to be sent to the
    instrument. This value mapping is such that, when inverted,
    ``on_val``/``off_val`` are mapped to boolean ``True``/``False``.
    """
    # Here are the lists of inputs which "reasonably" mean the same as
    # "on"/"off" (note that True/False values will be added below, and they
    # will always be added)
    ons_: Tuple[Union[str, bool], ...] = ('On',  'ON',  'on',  '1')
    offs_: Tuple[Union[str, bool], ...] = ('Off', 'OFF', 'off', '0')

    # The True/False values are added at the end of on/off inputs,
    # so that after inversion True/False will be the only remaining
    # keys in the inverted value mapping dictionary.
    # NOTE that using 1/0 integer values will also work implicitly
    # due to `hash(True) == hash(1)`/`hash(False) == hash(0)`,
    # hence there is no need for adding 1/0 values explicitly to
    # the list of `ons` and `offs` values.
    ons = ons_ + (True,)
    offs = offs_ + (False,)

    return OrderedDict([(on, on_val) for on in ons]
                       + [(off, off_val) for off in offs])


def abstractmethod(funcobj: Callable[..., Any]) -> Callable[..., Any]:
    """
    A decorator indicating abstract methods.

    This is heavily inspired by the decorator of the same name in
    the ABC standard library. But we make our own version because
    we actually want to allow the class with the abstract method to be
    instantiated and we will use this property to detect if the
    method is abstract and should be overwritten.
    """
    funcobj.__qcodes_is_abstract_method__ = True  # type: ignore[attr-defined]
    return funcobj


def _ruamel_importer() -> type:
    try:
        from ruamel_yaml import YAML
    except ImportError:
        try:
            from ruamel.yaml import YAML
        except ImportError:
            raise ImportError('No ruamel module found. Please install '
                              'either ruamel.yaml or ruamel_yaml.')
    return YAML


# YAML module to be imported. Resovles naming issues of YAML from pypi and
# anaconda
YAML = _ruamel_importer()


def get_qcodes_path(*subfolder: str) -> str:
    """
    Return full file path of the QCoDeS module. Additional arguments will be
    appended as subfolder.

    """
    import qcodes
    path = os.sep.join(qcodes.__file__.split(os.sep)[:-1])
    return os.path.join(path, *subfolder) + os.sep


def get_qcodes_user_path(*file_parts: str) -> str:
    """
    Get ``~/.qcodes`` path or if defined the path defined in the
    ``QCODES_USER_PATH`` environment variable.

    Returns:
        path to the user qcodes directory

    """
    path = os.environ.get(QCODES_USER_PATH_ENV,
                          os.path.join(Path.home(), '.qcodes'))
    os.makedirs(os.path.dirname(path), exist_ok=True)
    return os.path.join(path, *file_parts)


def checked_getattr(
    instance: Any,
    attribute: str,
    expected_type: Union[type, Tuple[type, ...]]
) -> Any:
    """
    Like ``getattr`` but raises type error if not of expected type.
    """
    attr: Any = getattr(instance, attribute)
    if not isinstance(attr, expected_type):
        raise TypeError()
    return attr



"""
Test module for snapshots of instruments and parameters
"""

from qcodes.tests.instrument_mocks import SnapShotTestInstrument
import pytest


@pytest.mark.parametrize("params,params_to_skip",
                         [(['v1', 'v2', 'v3', 'v4'], ['v1']),
                          (['v1', 'v2', 'v3', 'v4'], ['v2']),
                          (['v1', 'v2', 'v3', 'v4'], ['v3']),
                          (['v1', 'v2', 'v3', 'v4'], ['v4']),
                          (['v1', 'v2', 'v3', 'v4'], [])])
def test_snapshot_skip_params_update(request, params, params_to_skip):
    """
    Test that params_to_skip_update works as expected, in particular that only
    the parameters given by that variable are skipped.

    This test does not directly call snapshot_base, because this is the way a
    driver will work "in the wild"; the params_to_skip update are baked into
    the driver (in this case passed to the constructor) and the driver only
    calls :meth:`snapshot`
    """

    inst = SnapShotTestInstrument('snapshot_inst_1',
                                  params=params,
                                  params_to_skip=params_to_skip)
    request.addfinalizer(inst.close)

    assert list(inst._get_calls.values()) == [0, 0, 0, 0]

    inst.snapshot(update=False)

    assert list(inst._get_calls.values()) == [0, 0, 0, 0]

    inst.snapshot(update=True)

    expected_list = [1, 1, 1, 1]
    if params_to_skip:
        expected_list[params.index(params_to_skip[0])] = 0

    assert list(inst._get_calls.values()) == expected_list


@pytest.mark.parametrize("params,params_to_exclude",
                         [(['v1', 'v2', 'v3', 'v4'], ['v1']),
                          (['v1', 'v2', 'v3', 'v4'], ['v2']),
                          (['v1', 'v2', 'v3', 'v4'], ['v3']),
                          (['v1', 'v2', 'v3', 'v4'], ['v4']),
                          (['v1', 'v2', 'v3', 'v4'], ['v4']),
                          (['v1', 'v2', 'v3', 'v4'], ['v1', 'v2']),
                          (['v1', 'v2', 'v3', 'v4'], [])])
def test_snapshot_exclude_params(request, params, params_to_exclude):
    """
    Test that params_to_exclude works as expected, in particular that only
    the parameters given by that variable are excluded from the snapshot.
    This test does not directly call snapshot_base, because this is the way a
    driver will work "in the wild"; the params_to_exclude update are baked into
    the driver (in this case passed to the constructor) and the driver only
    calls :meth:`snapshot`
    """

    inst = SnapShotTestInstrument('snapshot_inst_2',
                                  params=params,
                                  params_to_skip=[])
    request.addfinalizer(inst.close)

    params.insert(0, "IDN")  # Is added by default to a instrument

    for param_exl in params_to_exclude:
        inst.parameters[param_exl].snapshot_exclude = True

    snap = inst.snapshot()
    snap_params = [x for x in snap["parameters"]]
    params = [x for x in params if x not in params_to_exclude]

    assert all(elem not in snap_params for elem in params_to_exclude), \
        f"Parameter(s) is not excluded from the snapshot. expected: " \
        f"{params}, actual: {snap_params}"\

    assert snap_params == params, f"Snapshot does not contain the expected " \
        f"parameters expected: {params}, actual: {snap_params}"



"""
Station objects - collect all the equipment you use to do an experiment.
"""


import importlib
import inspect
import itertools
import json
import logging
import os
import pkgutil
import warnings
from collections import deque
from contextlib import suppress
from copy import copy, deepcopy
from functools import partial
from io import StringIO
from pathlib import Path
from types import ModuleType
from typing import IO, Any, AnyStr
from typing import Deque as Tdeque
from typing import Dict, Iterable, List, Optional, Sequence, Tuple, Union, cast

import jsonschema
import ruamel.yaml

import qcodes
import qcodes.utils.validators as validators
from qcodes.instrument.base import Instrument, InstrumentBase
from qcodes.instrument.channel import ChannelList
from qcodes.instrument.parameter import (
    DelegateParameter,
    ManualParameter,
    Parameter,
    _BaseParameter,
)
from qcodes.monitor.monitor import Monitor
from qcodes.utils.deprecate import issue_deprecation_warning
from qcodes.utils.helpers import (
    YAML,
    DelegateAttributes,
    checked_getattr,
    get_qcodes_path,
    get_qcodes_user_path,
)
from qcodes.utils.metadata import Metadatable

log = logging.getLogger(__name__)

PARAMETER_ATTRIBUTES = ['label', 'unit', 'scale', 'inter_delay', 'post_delay',
                        'step', 'offset']

SCHEMA_TEMPLATE_PATH = os.path.join(
    get_qcodes_path('dist', 'schemas'),
    'station-template.schema.json')
SCHEMA_PATH = get_qcodes_user_path('schemas', 'station.schema.json')
STATION_YAML_EXT = '*.station.yaml'


def get_config_enable_forced_reconnect() -> bool:
    return qcodes.config["station"]["enable_forced_reconnect"]


def get_config_default_folder() -> Optional[str]:
    return qcodes.config["station"]["default_folder"]


def get_config_default_file() -> Optional[str]:
    return qcodes.config["station"]["default_file"]


def get_config_use_monitor() -> Optional[str]:
    return qcodes.config["station"]["use_monitor"]


ChannelOrInstrumentBase = Union[InstrumentBase, ChannelList]


class ValidationWarning(Warning):
    """Replacement for jsonschema.error.ValidationError as warning."""

    pass


class StationConfig(Dict[Any, Any]):
    def snapshot(self, update: bool = True) -> 'StationConfig':
        return self


class Station(Metadatable, DelegateAttributes):

    """
    A representation of the entire physical setup.

    Lists all the connected Components and the current default
    measurement (a list of actions).

    Args:
        *components: components to add immediately to the
            Station. Can be added later via ``self.add_component``.
        config_file: Path to YAML files to load the station config from.

            - If only one yaml file needed to be loaded, it should be passed
              as a string, e.g., '~/station.yaml'
            - If more than one yaml file needed, they should be supplied as
              a sequence of strings, e.g. ['~/station1.yaml', '~/station2.yaml']

        use_monitor: Should the QCoDeS monitor be activated for this station.
        default: Is this station the default?
        update_snapshot: Immediately update the snapshot of each
            component as it is added to the Station.

    """

    default: Optional['Station'] = None
    "Class attribute to store the default station."

    delegate_attr_dicts = ['components']
    """
    A list of names (strings) of dictionaries
    which are (or will be) attributes of ``self``,
    whose keys should be treated as attributes of ``self``.
    """

    config: Optional[StationConfig] = None
    """
    A user dict representing the YAML file that the station was loaded from"""

    def __init__(self, *components: Metadatable,
                 config_file: Optional[Union[str, Sequence[str]]] = None,
                 use_monitor: Optional[bool] = None, default: bool = True,
                 update_snapshot: bool = True, **kwargs: Any) -> None:
        super().__init__(**kwargs)

        # when a new station is defined, store it in a class variable
        # so it becomes the globally accessible default station.
        # You can still have multiple stations defined, but to use
        # other than the default one you must specify it explicitly.
        # If for some reason you want this new Station NOT to be the
        # default, just specify default=False
        if default:
            Station.default = self

        self.components: Dict[str, Metadatable] = {}
        for item in components:
            self.add_component(item, update_snapshot=update_snapshot)

        self.use_monitor = use_monitor

        self._added_methods: List[str] = []
        self._monitor_parameters: List[Parameter] = []

        if config_file is None:
            self.config_file = []
        elif isinstance(config_file, str):
            self.config_file = [config_file, ]
        else:
            self.config_file = list(config_file)

        self.load_config_files(*self.config_file)

    def snapshot_base(self, update: Optional[bool] = True,
                      params_to_skip_update: Optional[Sequence[str]] = None
                      ) -> Dict[Any, Any]:
        """
        State of the station as a JSON-compatible dictionary (everything that
        the custom JSON encoder class :class:`qcodes.utils.helpers.NumpyJSONEncoder`
        supports).

        Note: If the station contains an instrument that has already been
        closed, not only will it not be snapshotted, it will also be removed
        from the station during the execution of this function.

        Args:
            update: If ``True``, update the state by querying the
                all the children: f.ex. instruments, parameters,
                components, etc. If None only update if the state
                is known to be invalid.
                If ``False``, just use the latest
                values in memory and never update the state.
            params_to_skip_update: Not used.

        Returns:
            dict: Base snapshot.
        """
        snap: Dict[str, Any] = {
            'instruments': {},
            'parameters': {},
            'components': {},
            'config': self.config,
        }

        components_to_remove = []

        for name, itm in self.components.items():
            if isinstance(itm, Instrument):
                # instruments can be closed during the lifetime of the
                # station object, hence this 'if' allows to avoid
                # snapshotting instruments that are already closed
                if Instrument.is_valid(itm):
                    snap['instruments'][name] = itm.snapshot(update=update)
                else:
                    components_to_remove.append(name)
            elif isinstance(itm, (Parameter,
                                  ManualParameter
                                  )):
                if not itm.snapshot_exclude:
                    snap['parameters'][name] = itm.snapshot(update=update)
            else:
                snap['components'][name] = itm.snapshot(update=update)

        for c in components_to_remove:
            self.remove_component(c)

        return snap

    def add_component(self, component: Metadatable, name: Optional[str] = None,
                      update_snapshot: bool = True) -> str:
        """
        Record one component as part of this Station.

        Args:
            component: Components to add to the Station.
            name: Name of the component.
            update_snapshot: Immediately update the snapshot
                of each component as it is added to the Station.

        Returns:
            str: The name assigned this component, which may have been changed
                to make it unique among previously added components.
        """
        try:
            if not (isinstance(component, Parameter)
                    and component.snapshot_exclude):
                component.snapshot(update=update_snapshot)
        except:
            pass
        if name is None:
            name = getattr(component, "name", f"component{len(self.components)}")
        namestr = str(name)
        if namestr in self.components.keys():
            raise RuntimeError(
                f'Cannot add component "{namestr}", because a '
                'component of that name is already registered to the station')
        self.components[namestr] = component
        return namestr

    def remove_component(self, name: str) -> Optional[Metadatable]:
        """
        Remove a component with a given name from this Station.

        Args:
            name: Name of the component.

        Returns:
            The component that has been removed (this behavior is the same as
            for Python dictionaries).

        Raises:
            KeyError: If a component with the given name is not part of this
                station.
        """
        try:
            return self.components.pop(name)
        except KeyError as e:
            if name in str(e):
                raise KeyError(f'Component {name} is not part of the station')
            else:
                raise e

    # station['someitem'] and station.someitem are both
    # shortcuts to station.components['someitem']
    # (assuming 'someitem' doesn't have another meaning in Station)
    def __getitem__(self, key: str) -> Metadatable:
        """Shortcut to components dictionary."""
        return self.components[key]

    def close_all_registered_instruments(self) -> None:
        """
        Closes all instruments that are registered to this `Station`
        object by calling the :meth:`.base.Instrument.close`-method on
        each one.
        The instruments will be removed from the station and from the
        QCoDeS monitor.
        """
        for c in tuple(self.components.values()):
            if isinstance(c, Instrument):
                self.close_and_remove_instrument(c)

    @staticmethod
    def _get_config_file_path(
            filename: Optional[str] = None) -> Optional[str]:
        """
        Methods to get complete path of a provided file. If not able to find
        path then returns None.
        """
        filename = filename or get_config_default_file()
        if filename is None:
            return None
        search_list = [filename]
        if (not os.path.isabs(filename) and
                get_config_default_folder() is not None):
            config_folder = cast(str, get_config_default_folder())
            search_list += [os.path.join(config_folder, filename)]
        for p in search_list:
            if os.path.isfile(p):
                return p
        return None

    def load_config_file(self, filename: Optional[str] = None) -> None:
        """
        Loads a configuration from a YAML file. If `filename` is not specified
        the default file name from the qcodes configuration will be used.

        Loading of a configuration will update the snapshot of the station and
        make the instruments described in the config file available for
        instantiation with the :meth:`load_instrument` method.

        Additionally the shortcut methods ``load_<instrument_name>`` will be
        updated.
        """

        path = self._get_config_file_path(filename)

        if path is None:
            if filename is not None:
                raise FileNotFoundError(path)
            else:
                if get_config_default_file() is not None:
                    log.warning(
                        'Could not load default config for Station: \n'
                        f'File {get_config_default_file()} not found. \n'
                        'You can change the default config file in '
                        '`qcodesrc.json`.')
                return

        with open(path) as f:
            self.load_config(f)

    def load_config_files(self,
                          *filenames: str
                          ) -> None:
        """
        Loads configuration from multiple YAML files after merging them
        into one. If `filenames` are not specified the default file name from
        the qcodes configuration will be used.

        Loading of configuration will update the snapshot of the station and
        make the instruments described in the config files available for
        instantiation with the :meth:`load_instrument` method.

        Additionally the shortcut methods ``load_<instrument_name>`` will be
        updated.
        """
        if len(filenames) == 0:
            self.load_config_file()
        else:
            paths = list()
            for filename in filenames:
                assert isinstance(filename, str)
                path = self._get_config_file_path(filename)

                if path is None and filename is not None:
                    raise FileNotFoundError(path)

                paths.append(path)

            yamls = _merge_yamls(*paths)
            self.load_config(yamls)

    def load_config(self, config: Union[str, IO[AnyStr]]) -> None:
        """
        Loads a configuration from a supplied string or file/stream handle.
        The string or file/stream is expected to be YAML formatted

        Loading of a configuration will update the snapshot of the station and
        make the instruments described in the config file available for
        instantiation with the :meth:`load_instrument` method.

        Additionally the shortcut methods ``load_<instrument_name>`` will be
        updated.
        """

        def update_station_configuration_snapshot() -> None:
            self.config = StationConfig(self._config)

        def update_load_instrument_methods() -> None:
            #  create shortcut methods to instantiate instruments via
            # `load_<instrument_name>()` so that autocompletion can be used
            # first remove methods that have been added by a previous
            # :meth:`load_config_file` call
            while len(self._added_methods):
                delattr(self, self._added_methods.pop())

            # add shortcut methods
            for instrument_name in self._instrument_config.keys():
                method_name = f'load_{instrument_name}'
                if method_name.isidentifier():
                    setattr(self, method_name, partial(
                        self.load_instrument, identifier=instrument_name))
                    self._added_methods.append(method_name)
                else:
                    log.warning(f'Invalid identifier: '
                                f'for the instrument {instrument_name} no '
                                f'lazy loading method {method_name} could '
                                'be created in the Station.')

        # Load template schema, and thereby don't fail on instruments that are
        # not included in the user schema.
        yaml = YAML().load(config)
        with open(SCHEMA_TEMPLATE_PATH) as f:
            schema = json.load(f)
        try:
            jsonschema.validate(yaml, schema)
        except jsonschema.exceptions.ValidationError as e:
            message = str(e)
            warnings.warn(message, ValidationWarning)

        self._config = yaml

        self._instrument_config = self._config['instruments']
        update_station_configuration_snapshot()
        update_load_instrument_methods()

    def close_and_remove_instrument(self,
                                    instrument: Union[Instrument, str]
                                    ) -> None:
        """
        Safely close instrument and remove from station and monitor list.
        """
        # remove parameters related to this instrument from the
        # monitor list
        if isinstance(instrument, str):
            instrument = Instrument.find_instrument(instrument)

        self._monitor_parameters = [v for v in self._monitor_parameters
                                    if v.root_instrument is not instrument]
        # remove instrument from station snapshot
        self.remove_component(instrument.name)
        # del will remove weakref and close the instrument
        instrument.close()
        del instrument

    def load_instrument(self, identifier: str,
                        revive_instance: bool = False,
                        **kwargs: Any) -> Instrument:
        """
        Creates an :class:`~.Instrument` instance as described by the
        loaded configuration file.

        Args:
            identifier: The identfying string that is looked up in the yaml
                configuration file, which identifies the instrument to be added.
            revive_instance: If ``True``, try to return an instrument with the
                specified name instead of closing it and creating a new one.
            **kwargs: Additional keyword arguments that get passed on to the
                ``__init__``-method of the instrument to be added.
        """
        # try to revive the instrument
        if revive_instance and Instrument.exist(identifier):
            return Instrument.find_instrument(identifier)

        # load file
        # try to reload file on every call. This makes script execution a
        # little slower but makes the overall workflow more convenient.
        self.load_config_files(*self.config_file)


        # load from config
        if identifier not in self._instrument_config.keys():
            raise RuntimeError(f'Instrument {identifier} not found in '
                               'instrument config file')
        instr_cfg = self._instrument_config[identifier]

        # TODO: add validation of config for better verbose errors:

        # check if instrument is already defined and close connection
        if instr_cfg.get('enable_forced_reconnect',
                         get_config_enable_forced_reconnect()):
            with suppress(KeyError):
                self.close_and_remove_instrument(identifier)

        # instantiate instrument
        init_kwargs = instr_cfg.get('init', {})
        # somebody might have a empty init section in the config
        init_kwargs = {} if init_kwargs is None else init_kwargs
        if 'address' in instr_cfg:
            init_kwargs['address'] = instr_cfg['address']
        if 'port' in instr_cfg:
            init_kwargs['port'] = instr_cfg['port']
        # make explicitly passed arguments overide the ones from the config
        # file.
        # We are mutating the dict below
        # so make a copy to ensure that any changes
        # does not leek into the station config object
        # specifically we may be passing non pickleable
        # instrument instances via kwargs
        instr_kwargs = deepcopy(init_kwargs)
        instr_kwargs.update(kwargs)
        name = instr_kwargs.pop('name', identifier)

        if 'driver' in instr_cfg:
            issue_deprecation_warning(
                'use of the "driver"-keyword in the station '
                'configuration file',
                alternative='the "type"-keyword instead, prepending the '
                'driver value'
                ' to it')
            module_name = instr_cfg['driver']
            instr_class_name = instr_cfg['type']
        else:
            module_name = '.'.join(instr_cfg['type'].split('.')[:-1])
            instr_class_name = instr_cfg['type'].split('.')[-1]
        module = importlib.import_module(module_name)
        instr_class = getattr(module, instr_class_name)
        instr = instr_class(name=name, **instr_kwargs)

        def resolve_instrument_identifier(
            instrument: ChannelOrInstrumentBase,
            identifier: str
        ) -> ChannelOrInstrumentBase:
            """
            Get the instrument, channel or channel_list described by a nested
            string.

            E.g: 'dac.ch1' will return the instance of ch1.
            """
            try:
                for level in identifier.split('.'):
                    instrument = checked_getattr(
                        instrument, level,
                        (InstrumentBase, ChannelList))
            except TypeError:
                raise RuntimeError(
                    f'Cannot resolve `{level}` in {identifier} to an '
                    f'instrument/channel for base instrument '
                    f'{instrument!r}.')
            return instrument

        def resolve_parameter_identifier(
            instrument: ChannelOrInstrumentBase,
            identifier: str
        ) -> _BaseParameter:
            parts = identifier.split('.')
            if len(parts) > 1:
                instrument = resolve_instrument_identifier(
                    instrument,
                    '.'.join(parts[:-1]))
            try:
                return checked_getattr(instrument, parts[-1], _BaseParameter)
            except TypeError:
                raise RuntimeError(
                    f'Cannot resolve parameter identifier `{identifier}` to '
                    f'a parameter on instrument {instrument!r}.')

        def setup_parameter_from_dict(
            parameter: _BaseParameter,
            options: Dict[str, Any]
        ) -> None:
            for attr, val in options.items():
                if attr in PARAMETER_ATTRIBUTES:
                    # set the attributes of the parameter, that map 1 to 1
                    setattr(parameter, attr, val)
                # extra attributes that need parsing
                elif attr == 'limits':
                    if isinstance(val, str):
                        issue_deprecation_warning(
                            (
                                "use of a comma separated string for the limits "
                                "keyword"
                            ),
                            alternative='an array like "[lower_lim, upper_lim]"',
                        )
                        lower, upper = (float(x) for x in val.split(","))
                    else:
                        lower, upper = val
                    parameter.vals = validators.Numbers(lower, upper)
                elif attr == 'monitor' and val is True:
                    if isinstance(parameter, Parameter):
                        self._monitor_parameters.append(parameter)
                    else:
                        raise RuntimeError(f"Trying to add {parameter} to "
                                           f"monitored parameters. But it's "
                                           f"not a Parameter but a"
                                           f" {type(parameter)}")
                elif attr == 'alias':
                    setattr(parameter.instrument, val, parameter)
                elif attr == 'initial_value':
                    # skip value attribute so that it gets set last
                    # when everything else has been set up
                    pass
                else:
                    log.warning(f'Attribute {attr} not recognized when '
                                f'instatiating parameter \"{parameter.name}\"')
            if 'initial_value' in options:
                parameter.set(options['initial_value'])

        def add_parameter_from_dict(
            instr: InstrumentBase,
            name: str,
            options: Dict[str, Any]
        ) -> None:
            # keep the original dictionray intact for snapshot
            options = copy(options)
            param_type: type = _BaseParameter
            kwargs = {}
            if 'source' in options:
                param_type = DelegateParameter
                kwargs['source'] = resolve_parameter_identifier(
                    instr.root_instrument,
                    options['source'])
                options.pop('source')
            instr.add_parameter(name, param_type, **kwargs)
            setup_parameter_from_dict(instr.parameters[name], options)

        def update_monitor() -> None:
            if ((self.use_monitor is None and get_config_use_monitor())
                    or self.use_monitor):
                # restart Monitor
                Monitor(*self._monitor_parameters)

        for name, options in instr_cfg.get('parameters', {}).items():
            parameter = resolve_parameter_identifier(instr, name)
            setup_parameter_from_dict(parameter, options)
        for name, options in instr_cfg.get('add_parameters', {}).items():
            parts = name.split('.')
            local_instr = (
                instr if len(parts) < 2 else
                resolve_instrument_identifier(instr, '.'.join(parts[:-1])))
            add_parameter_from_dict(local_instr, parts[-1], options)
        self.add_component(instr)
        update_monitor()
        return instr

    def load_all_instruments(self,
                             only_names: Optional[Iterable[str]] = None,
                             only_types: Optional[Iterable[str]] = None,
                             ) -> Tuple[str, ...]:
        """
        Load all instruments specified in the loaded YAML station
        configuration.

        Optionally, the instruments to be loaded can be filtered by their
        names or types, use ``only_names`` and ``only_types``
        arguments for that. It is an error to supply both ``only_names``
        and ``only_types``.

        Args:
            only_names: List of instrument names to load from the config.
                If left as None, then all instruments are loaded.
            only_types: List of instrument types e.g. the class names
                of the instruments to load. If left as None, then all
                instruments are loaded.

        Returns:
            The names of the loaded instruments
        """
        config = self.config
        if config is None:
            raise ValueError("Station has no config")

        instrument_names_to_load = set()

        if only_names is None and only_types is None:
            instrument_names_to_load = set(config["instruments"].keys())
        elif only_types is None and only_names is not None:
            instrument_names_to_load = set(only_names)
        elif only_types is not None and only_names is None:
            for inst_name, inst_dict in config["instruments"].items():
                if "driver" in inst_dict.keys():
                    # fallback for old format where type was used
                    # together with the driver key.
                    inst_type = inst_dict["type"]
                else:
                    inst_type = inst_dict["type"].split(".")[-1]
                if inst_type in only_types:
                    instrument_names_to_load.add(inst_name)
        else:
            raise ValueError(
                "It is an error to supply both ``only_names`` "
                "and ``only_types`` arguments."
            )

        for instrument in instrument_names_to_load:
            self.load_instrument(instrument)

        return tuple(instrument_names_to_load)


def update_config_schema(
    additional_instrument_modules: Optional[List[ModuleType]] = None
) -> None:
    """Update the json schema file 'station.schema.json'.

    Args:
        additional_instrument_modules: python modules that contain
            :class:`qcodes.instrument.base.InstrumentBase` definitions
            (and subclasses thereof) to be included as
            values for instrument definition in th station definition
            yaml files.

    """

    def instrument_names_from_module(module: ModuleType) -> Tuple[str, ...]:
        submodules = list(pkgutil.walk_packages(
            module.__path__,  # type: ignore[attr-defined]  # mypy issue #1422
            module.__name__ + '.'))
        res = set()
        for s in submodules:
            with suppress(Exception):
                ms = inspect.getmembers(
                    importlib.import_module(s.name),
                    inspect.isclass)
            new_members = [
                f"{instr[1].__module__}.{instr[1].__name__}"
                for instr in ms
                if (issubclass(instr[1], InstrumentBase) and
                    instr[1].__module__.startswith(module.__name__))
            ]
            res.update(new_members)
        return tuple(res)

    def update_schema_file(
        template_path: str,
        output_path: str,
        instrument_names: Tuple[str, ...]
    ) -> None:
        with open(template_path, 'r+') as f:
            data = json.load(f)
        data['definitions']['instruments']['enum'] = instrument_names
        if os.path.exists(output_path):
            os.remove(output_path)
        os.makedirs(os.path.dirname(output_path), exist_ok=True)
        with open(output_path, 'w') as f:
            json.dump(data, f, indent=4)

    additional_instrument_modules = additional_instrument_modules or []
    update_schema_file(
        template_path=SCHEMA_TEMPLATE_PATH,
        output_path=SCHEMA_PATH,
        instrument_names=tuple(itertools.chain.from_iterable(
            instrument_names_from_module(m)
            for m in set(
                [qcodes.instrument_drivers] + additional_instrument_modules)
        ))
    )


def _merge_yamls(*yamls: Union[str, Path]) -> str:
    """
    Merge multiple station yamls files into one and stores it in the memory.

    Args:
        yamls: string or Path to yaml files separated by comma.
    Returns:
        Full yaml file stored in the memory.
    """

    if len(yamls) == 1:
        with open(yamls[0]) as file:
            content = file.read()
        return content

    top_key = "instruments"
    yaml = ruamel.yaml.YAML()

    deq: Tdeque[Any] = deque()

    # Load the yaml files and add to deque in reverse entry order
    for filepath in yamls[::-1]:
        with open(filepath) as file_pointer:
            deq.append(yaml.load(file_pointer))

    # Add the top key entries from filepath n to filepath n-1 to
    # ... filepath 1.
    while len(deq) > 1:
        data2, data1 = deq[0], deq[1]
        for entry in data2[top_key]:
            if entry not in data1[top_key].keys():
                data1[top_key].update({entry: data2[top_key][entry]})
            else:
                raise KeyError(
                    f"duplicate key `{entry}` detected among files:"
                    f"{ ','.join(map(str, yamls))}"
                )
        deq.popleft()

    with StringIO() as merged_yaml_stream:
        yaml.dump(data1, merged_yaml_stream)
        merged_yaml = merged_yaml_stream.getvalue()
    return merged_yaml



import pytest

from qcodes.instrument.parameter import InstrumentRefParameter
from qcodes.tests.instrument_mocks import DummyInstrument


@pytest.fixture()
def instrument_a():
    a = DummyInstrument('dummy_holder')
    try:
        yield a
    finally:
        a.close()


@pytest.fixture()
def instrument_d():
    d = DummyInstrument('dummy')
    try:
        yield d
    finally:
        d.close()


def test_get_instr(instrument_a, instrument_d):
    instrument_a.add_parameter('test', parameter_class=InstrumentRefParameter)

    instrument_a.test.set(instrument_d.name)

    assert instrument_a.test.get() == instrument_d.name
    assert instrument_a.test.get_instr() == instrument_d



"""
This module provides helper objects for Alazar driver class. The objects allow
to hide realted pieces of logic into a form of "submodule" (analogious to
:meth:`.InstrumentBase.add_submodule`) that can be included into the driver
class in some way.
"""


from .ats_api import AlazarATSAPI
from .constants import Capability


class CapabilityHelper():
    """
    A helper class providing convenient methods for various useful
    'query capability' (``AlazarQueryCapability``) calls for a given
    Alazar board.

    Most common capabilities are enumerated in :attr:`.CAPABILITIES`.

    For frequently used capabilities, dedicated convenience ``query_<...>()``
    methods are available.

    Args:
        api: Instance of Alazar ATS API class
        handle: Handle of a specific board (from ``AlazarGetBoardBySystemId``)
    """

    CAPABILITIES = Capability

    def __init__(self, api: AlazarATSAPI, handle: int):
        self._api = api
        self._handle = handle

    def query(self, capability: int) -> int:
        """Query the given 'capability' of the board"""
        return self._api.query_capability_(self._handle, capability)

    # Convenience and specific methods

    def query_serial(self) -> str:
        return str(self.query(self.CAPABILITIES.GET_SERIAL_NUMBER))

    def query_latest_calibration(self) -> str:
        """Query latest calibration date in '12-34-56' format"""
        # ``date_int`` is a decimal number with the format DDMMYY where
        # DD is 1-31, MM is 1-12, and YY is 00-99 from 2000.
        date_int = self.query(self.CAPABILITIES.GET_LATEST_CAL_DATE)
        date_str = str(date_int)
        date = date_str[0:2] + "-" + date_str[2:4] + "-" + date_str[4:6]
        return date

    def query_memory_size(self) -> int:
        """Query board memory size in samples"""
        return self.query(self.CAPABILITIES.MEMORY_SIZE)

    def query_asopc_type(self) -> int:
        return self.query(self.CAPABILITIES.ASOPC_TYPE)

    def query_pcie_link_speed(self) -> float:
        """Query PCIE link speed in GB/s"""
        # See the ATS-SDK programmer's guide about the encoding
        # of the PCIE link speed.
        link_speed_int = self.query(self.CAPABILITIES.GET_PCIE_LINK_SPEED)
        link_speed = link_speed_int * 2.5 / 10
        return link_speed

    def query_pcie_link_width(self) -> int:
        """Query PCIE link width"""
        return self.query(self.CAPABILITIES.GET_PCIE_LINK_WIDTH)

    def query_firmware_version(self) -> str:
        """
        Query firmware version in "<major>.<minor>" format

        The firmware version reported should match the version number of
        downloadable fw files from AlazarTech. But note that the firmware
        version has often been found to be incorrect for several firmware
        versions. At the time of writing it is known to be correct for the
        9360 (v 21.07) and 9373 (v 30.04) but incorrect for several earlier
        versions. In Alazar DSO this is reported as FPGA Version.
        """
        asopc_type = self.query_asopc_type()
        # AlazarTech has confirmed in a support mail that this
        # is the way to get the firmware version
        firmware_major = (asopc_type >> 16) & 0xff
        firmware_minor = (asopc_type >> 24) & 0xf
        # firmware_minor above does not contain any prefixed zeros
        # but the minor version is always 2 digits.
        firmware_version = f'{firmware_major}.{firmware_minor:02d}'
        return firmware_version



from collections.abc import Iterable

import hypothesis.strategies as hst
from hypothesis import given, event, settings
import numpy as np

from qcodes.instrument.parameter import Parameter


def test_scale_raw_value():
    p = Parameter(name='test_scale_raw_value', set_cmd=None)
    p(42)
    assert p.raw_value == 42

    p.scale = 2
    assert p.raw_value == 42  # No set/get cmd performed
    assert p() == 21

    p(10)
    assert p.raw_value == 20
    assert p() == 10


# There are a number different scenarios for testing a parameter with scale
# and offset. Therefore a custom strategy for generating test parameters
# is implemented here. The possible cases are:
# for getting and setting a parameter: values can be
#    scalar:
#        offset and scale can be scalars
# for getting only:
#    array:
#        offset and scale can be scalars or arrays(of same legnth as values)
#        independently

# define shorthands for strategies
TestFloats = hst.floats(min_value=-1e40, max_value=1e40).filter(lambda x: x != 0)
SharedSize = hst.shared(hst.integers(min_value=1, max_value=100), key='shared_size')
ValuesScalar = hst.shared(hst.booleans(), key='values_scalar')


# the following test stra
@hst.composite
def iterable_or_number(draw, values, size, values_scalar, is_values):
    if draw(values_scalar):
        # if parameter values are scalar,
        # return scalar for values and scale/offset
        return draw(values)
    elif is_values:
        # if parameter values are not scalar and parameter values are requested
        # return a list of values of the given size
        return draw(hst.lists(values, min_size=draw(size), max_size=draw(size)))
    else:
        # if parameter values are not scalar and scale/offset are requested
        # make a random choice whether to return a list of the same size as
        # the values or a simple scalar
        if draw(hst.booleans()):
            return draw(hst.lists(values, min_size=draw(size), max_size=draw(size)))
        else:
            return draw(values)


@settings(max_examples=500)  # default:100 increased
@given(values=iterable_or_number(TestFloats, SharedSize, ValuesScalar, True),
       offsets=iterable_or_number(TestFloats, SharedSize, ValuesScalar, False),
       scales=iterable_or_number(TestFloats, SharedSize, ValuesScalar, False))
def test_scale_and_offset_raw_value_iterable(values, offsets, scales):
    p = Parameter(name='test_scale_and_offset_raw_value', set_cmd=None)

    # test that scale and offset does not change the default behaviour
    p(values)
    assert p.raw_value == values

    # test setting scale and offset does not change anything
    p.scale = scales
    p.offset = offsets
    assert p.raw_value == values

    np_values = np.array(values)
    np_offsets = np.array(offsets)
    np_scales = np.array(scales)
    np_get_values = np.array(p())
    # No set/get cmd performed
    np.testing.assert_allclose(np_get_values, (np_values - np_offsets) / np_scales)

    # test set, only for scalar values
    if not isinstance(values, Iterable):
        p(values)
        # No set/get cmd performed
        np.testing.assert_allclose(np.array(p.raw_value),
                                   np_values * np_scales + np_offsets)

        # testing conversion back and forth
        p(values)
        np_get_values = np.array(p())
        # No set/get cmd performed
        np.testing.assert_allclose(np_get_values, np_values)

    # adding statistics
    if isinstance(offsets, Iterable):
        event('Offset is array')
    if isinstance(scales, Iterable):
        event('Scale is array')
    if isinstance(values, Iterable):
        event('Value is array')
    if isinstance(scales, Iterable) and isinstance(offsets, Iterable):
        event('Scale is array and also offset')
    if isinstance(scales, Iterable) and not isinstance(offsets, Iterable):
        event('Scale is array but not offset')


@settings(max_examples=300)
@given(
    values=iterable_or_number(
        TestFloats, SharedSize, ValuesScalar, True),
    offsets=iterable_or_number(
        TestFloats, SharedSize, ValuesScalar, False),
    scales=iterable_or_number(
        TestFloats, SharedSize, ValuesScalar, False))
def test_scale_and_offset_raw_value_iterable_for_set_cache(values,
                                                           offsets,
                                                           scales):
    p = Parameter(name='test_scale_and_offset_raw_value', set_cmd=None)

    # test that scale and offset does not change the default behaviour
    p.cache.set(values)
    assert p.raw_value == values

    # test setting scale and offset does not change anything
    p.scale = scales
    p.offset = offsets
    assert p.raw_value == values

    np_values = np.array(values)
    np_offsets = np.array(offsets)
    np_scales = np.array(scales)
    np_get_latest_values = np.array(p.get_latest())
    # Without a call to ``get``, ``get_latest`` will just return old
    # cached values without applying the set scale and offset
    np.testing.assert_allclose(np_get_latest_values, np_values)
    np_get_values = np.array(p.get())
    # Now that ``get`` is called, the returned values are the result of
    # application of the scale and offset. Obviously, calling
    # ``get_latest`` now will also return the values with the applied
    # scale and offset
    np.testing.assert_allclose(np_get_values,
                               (np_values - np_offsets) / np_scales)
    np_get_latest_values_after_get = np.array(p.get_latest())
    np.testing.assert_allclose(np_get_latest_values_after_get,
                               (np_values - np_offsets) / np_scales)

    # test ``cache.set`` for scalar values
    if not isinstance(values, Iterable):
        p.cache.set(values)
        np.testing.assert_allclose(np.array(p.raw_value),
                                   np_values * np_scales + np_offsets)
        # No set/get cmd performed

        # testing conversion back and forth
        p.cache.set(values)
        np_get_latest_values = np.array(p.get_latest())
        # No set/get cmd performed
        np.testing.assert_allclose(np_get_latest_values, np_values)

    # adding statistics
    if isinstance(offsets, Iterable):
        event('Offset is array')
    if isinstance(scales, Iterable):
        event('Scale is array')
    if isinstance(values, Iterable):
        event('Value is array')
    if isinstance(scales, Iterable) and isinstance(offsets, Iterable):
        event('Scale is array and also offset')
    if isinstance(scales, Iterable) and not isinstance(offsets, Iterable):
        event('Scale is array but not offset')


def test_numpy_array_valued_parameter_preserves_type_if_scale_and_offset_are_set():

    def rands():
        return np.random.randn(5)

    param = Parameter(name='test_param',
                       set_cmd=None,
                       get_cmd=rands)

    param.scale = 10
    param.offset = 7

    values = param()

    assert isinstance(values, np.ndarray)


def test_setting_numpy_array_valued_param_if_scale_and_offset_are_not_none():

    param = Parameter(name='test_param',
                      set_cmd=None,
                      get_cmd=None)

    values = np.array([1, 2, 3, 4, 5])

    param.scale = 100
    param.offset = 10

    param(values)

    assert isinstance(param.raw_value, np.ndarray)



"""
Useful collections of types used around QCoDeS
"""
from typing import Tuple, Union

import numpy as np

numpy_concrete_ints = (np.int8, np.int16, np.int32, np.int64,
                       np.uint8, np.uint16, np.uint32, np.uint64)
"""
Integer types with fixed sizes.
"""
numpy_c_ints = (np.uintp, np.uintc, np.intp, np.intc,
                np.short, np.byte, np.ushort, np.ubyte,
                np.longlong, np.ulonglong)
"""
Integer types that matches C types.
"""
numpy_non_concrete_ints_instantiable = (np.int_, np.uint)
"""
Default integer types. The size may be platform dependent.
"""

numpy_ints: Tuple[type, ...] = (
        numpy_concrete_ints +
        numpy_c_ints +
        numpy_non_concrete_ints_instantiable
)
"""
All numpy integer types
"""


numpy_concrete_floats = (np.float16, np.float32, np.float64)
"""
Floating point types with fixed sizes.
"""
numpy_c_floats = (np.half, np.single, np.double)
"""
Floating point types that matches C types.
"""
numpy_non_concrete_floats_instantiable = (np.float_, )
"""
Default floating point types. The size may be platform dependent.
"""

numpy_floats: Tuple[type, ...] = (
        numpy_concrete_floats +
        numpy_c_floats +
        numpy_non_concrete_floats_instantiable
)
"""
All numpy float types
"""

numpy_concrete_complex = (np.complex64, np.complex128)
"""
Complex types with fixed sizes.
"""
numpy_c_complex = (np.csingle, np.cdouble)
"""
Complex types that matches C types.
"""
numpy_non_concrete_complex_instantiable = (np.complex_, )
"""
Default complex types. The size may be platform dependent.
"""

numpy_complex = (
        numpy_concrete_complex +
        numpy_c_complex +
        numpy_non_concrete_complex_instantiable
)
"""
All numpy complex types
"""

concrete_complex_types = numpy_concrete_complex + (complex,)
complex_types = (
        numpy_concrete_complex +
        numpy_non_concrete_complex_instantiable +
        (complex,)
)

# These are the same types as a above unfortunately there does not seem to be
# a good way to convert a tuple of types to a Union
complex_type_union = Union[np.complex64, np.complex128,
                           np.complex_, np.complexfloating, complex]



from typing import Dict, Any, Optional, Tuple

from qcodes.instrument.base import Instrument
from qcodes.instrument.parameter import MultiParameter, Parameter, ParamRawDataType
from qcodes.utils.validators import Enum, Bool


class CurrentParameter(MultiParameter):
    """
    Voltage measurement via an Ithaco preamp and converting volt to current.

    To be used when you feed a current into the Ithaco, send the Ithaco's
    output voltage to a lockin or other voltage amplifier, and you have
    the voltage reading from that amplifier as a qcodes parameter.

    ``CurrentParameter.get()`` returns ``(volt_raw, curr)``

    Args:
        measured_param: a gettable parameter returning the
            voltage read from the Ithaco output.

        c_amp_ins: an Ithaco instance where you manually
            maintain the present settings of the real Ithaco amp.

            Note: it should be possible to use other current preamps, if they
            define parameters ``sens`` (sensitivity, in A/V), ``sens_factor``
            (an additional gain) and ``invert`` (bool, output is inverted)

        name: the name of the current output. Default 'curr'.
            Also used as the name of the whole parameter.
    """
    def __init__(self,
                 measured_param: Parameter,
                 c_amp_ins: "Ithaco_1211",
                 name: str = 'curr'):
        p_name = measured_param.name

        super().__init__(name=name,
                         names=(p_name+'_raw', name),
                         shapes=((), ()),
                         setpoints=((), ()),
                         instrument=c_amp_ins,
                         snapshot_value=True)

        self._measured_param = measured_param

        p_label = getattr(measured_param, 'label', None)
        p_unit = getattr(measured_param, 'unit', None)

        self.labels = (p_label, 'Current')
        self.units = (p_unit, 'A')

    def get_raw(self) -> Tuple[ParamRawDataType, ...]:
        assert isinstance(self.instrument, Ithaco_1211)
        volt = self._measured_param.get()
        current = (self.instrument.sens.get() *
                   self.instrument.sens_factor.get()) * volt

        if self.instrument.invert.get():
            current *= -1

        value = (volt, current)
        return value


class Ithaco_1211(Instrument):
    """
    This is the qcodes driver for the Ithaco 1211 Current-preamplifier.

    This is a virtual driver only and will not talk to your instrument.
    """
    def __init__(self, name: str, **kwargs: Any):
        super().__init__(name, **kwargs)

        self.add_parameter('sens',
                           initial_value=1e-8,
                           label='Sensitivity',
                           unit='A/V',
                           get_cmd=None, set_cmd=None,
                           vals=Enum(1e-11, 1e-10, 1e-09, 1e-08, 1e-07,
                                     1e-06, 1e-05, 1e-4, 1e-3))

        self.add_parameter('invert',
                           initial_value=True,
                           label='Inverted output',
                           get_cmd=None, set_cmd=None,
                           vals=Bool())

        self.add_parameter('sens_factor',
                           initial_value=1,
                           label='Sensitivity factor',
                           unit=None,
                           get_cmd=None, set_cmd=None,
                           vals=Enum(0.1, 1, 10))

        self.add_parameter('suppression',
                           initial_value=1e-7,
                           label='Suppression',
                           unit='A',
                           get_cmd=None, set_cmd=None,
                           vals=Enum(1e-10, 1e-09, 1e-08, 1e-07, 1e-06,
                                     1e-05, 1e-4, 1e-3))

        self.add_parameter('risetime',
                           initial_value=0.3,
                           label='Rise Time',
                           unit='msec',
                           get_cmd=None, set_cmd=None,
                           vals=Enum(0.01, 0.03, 0.1, 0.3, 1, 3, 10, 30,
                                     100, 300, 1000))

    def get_idn(self) -> Dict[str, Optional[str]]:
        vendor = 'Ithaco (DL Instruments)'
        model = '1211'
        serial = None
        firmware = None
        return {'vendor': vendor, 'model': model,
                'serial': serial, 'firmware': firmware}



import importlib
import logging
from collections import abc
from functools import partial
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Sequence,
    Type,
    Union,
)

from qcodes.instrument.base import InstrumentBase
from qcodes.instrument.channel import InstrumentChannel
from qcodes.instrument.delegate.grouped_parameter import (
    DelegateGroup,
    DelegateGroupParameter,
    GroupedParameter,
)
from qcodes.instrument.parameter import Parameter
from qcodes.station import Station

_log = logging.getLogger(__name__)


class DelegateInstrument(InstrumentBase):
    """DelegateInstrument is an instrument driver with one or more
    parameters that connect to instrument parameters.

    Example usage in instrument YAML:

    .. code-block:: yaml

        field:
            type: qcodes.instrument.delegate.DelegateInstrument
            init:
            parameters:
                X:
                    - field_X.field
                ramp_rate:
                    - field_X.ramp_rate
            channels:
                gate_1: dac.ch01
            set_initial_values_on_load: true
            initial_values:
                ramp_rate: 0.02
            setters:
                X:
                    method: field_X.set_field
                    block: false
            units:
                X: T
                ramp_rate: T/min

    this will generate an instrument named "field" with methods:

    .. code-block:: python

        field.X()
        field.ramp_rate()

    that are delegate parameters for:

    .. code-block:: python

        field_X.field()
        field_X.ramp_rate()

    Additionally, this will set ``field_X.ramp_rate(0.02)``` on load and
    override the ``field.X.set()`` method with

    .. code-block:: python

        field_X.set_field(value, block=False),

    as opposed to ``field.X.field.set()`` which ramps with ``block=True``.


    Args:
        name: Instrument name
        station: Station containing the real instrument that is used to get
            the endpoint parameters.
        parameters: A mapping from the name of a parameter to the sequence
            of source parameters that it points to.
        channels: A mapping from the name of an instrument channel to either
            the channel it emulates or a mapping of keyworded input parameters
            of a custom channel wrapper class. This custom channel wrapper
            class needs to be specified under the `type` keyword. Each
            channel can have its own `type` if required. A single type for all
            channels can also be specified.
        initial_values: Default values to set on the delegate instrument's
            parameters. Defaults to None (no initial values are specified or
            set).
        set_initial_values_on_load: Flag to set initial values when the
            instrument is loaded. Defaults to False.
        setters: Optional setter methods to use instead of calling the
            ``.set()`` method on the endpoint parameters. Defaults to None.
        units: Optional units to set for parameters.
        metadata: Optional metadata to pass to instrument. Defaults to None.
    """

    param_cls = DelegateGroupParameter

    def __init__(
        self,
        name: str,
        station: Station,
        parameters: Optional[
            Union[Mapping[str, Sequence[str]], Mapping[str, str]]
        ] = None,
        channels: Optional[
            Union[Mapping[str, Mapping[str, Any]], Mapping[str, str]]
        ] = None,
        initial_values: Optional[Mapping[str, Any]] = None,
        set_initial_values_on_load: bool = False,
        setters: Optional[Mapping[str, MutableMapping[str, Any]]] = None,
        units: Optional[Mapping[str, str]] = None,
        metadata: Optional[Mapping[Any, Any]] = None,
    ):
        super().__init__(name=name, metadata=metadata)
        if parameters is not None:
            self._create_and_add_parameters(
                station=station,
                parameters=parameters,
                setters=setters or {},
                units=units or {},
            )

        if channels is not None:
            self._create_and_add_channels(
                station=station,
                channels=channels,
            )

        self._initial_values = initial_values or {}
        if set_initial_values_on_load:
            self.set_initial_values()

    @staticmethod
    def parse_instrument_path(
        parent: Union[Station, InstrumentBase],
        path: str,
    ) -> Any:
        """Parse a string path and return the object relative to a station or
        instrument, e.g. "my_instrument.my_param" returns
        station.my_instrument.my_param

        Args:
            parent: Measurement station
            path: Relative path to parse
        """
        def _parse_path(parent: Any, elem: Sequence[str]) -> Any:
            child = getattr(parent, elem[0])
            if len(elem) == 1:
                return child
            return _parse_path(child, elem[1:])

        return _parse_path(parent, path.split("."))

    def set_initial_values(self, dry_run: bool = False) -> None:
        """Set parameter initial values on delegate instrument

        Args:
            dry_run: Dry run to test if defaults are set correctly.
                Defaults to False.
        """
        _log.debug(f"Setting default values: {self._initial_values}")
        for path, value in self._initial_values.items():
            param = self.parse_instrument_path(parent=self, path=path)
            msg = f"Setting parameter {self.name}.{path} to {value}."
            if not dry_run:
                _log.debug(msg)
                if hasattr(param, "set"):
                    param.set(value)
                else:
                    _log.debug("No set method found, trying to assign value.")
                    if "." in path:
                        name = path.split(".")[-1]
                        parent_path = ".".join(path.split(".")[:-1])
                        parent = self.parse_instrument_path(
                            parent=self,
                            path=parent_path
                        )
                    else:
                        parent, name = self, path
                    setattr(parent, name, value)
            else:
                print(f"Dry run: {msg}")

    def _create_and_add_parameters(
        self,
        station: Station,
        parameters: Union[Mapping[str, Sequence[str]], Mapping[str, str]],
        setters: Mapping[str, MutableMapping[str, Any]],
        units: Mapping[str, str],
    ) -> None:
        """Add parameters to delegate instrument based on specified aliases,
        endpoints and setter methods"""
        for param_name, paths in parameters.items():
            if isinstance(paths, str):
                path_list: Sequence[str] = [paths]

            elif isinstance(paths, abc.Sequence):
                path_list = paths
            else:
                raise ValueError(
                    "Parameter paths should be either a string or Sequence of \
                        strings."
                )

            self._create_and_add_parameter(
                group_name=param_name,
                station=station,
                paths=path_list,
                setter=setters.get(param_name),
                unit=units.get(param_name),
            )

    @staticmethod
    def _parameter_names(parameters: Sequence[Parameter]) -> List[str]:
        """Get the endpoint names"""
        parameter_names = [_e.name for _e in parameters]
        if len(parameter_names) != len(set(parameter_names)):
            parameter_names = [
                f"{_e}{n}" for n, _e in enumerate(parameter_names)
            ]
        return parameter_names

    def _add_parameter(
        self,
        group_name: str,
        name: str,
        source: Parameter,
    ) -> DelegateGroupParameter:
        param_name = f"{group_name}_{name}"
        if param_name in self.parameters:
            raise KeyError(f"Duplicate parameter name {param_name} on {self.name}")

        self.add_parameter(
            parameter_class=self.param_cls,
            name=param_name,
            source=source,
        )
        param = self.parameters[param_name]
        assert isinstance(param, DelegateGroupParameter)

        return param

    def _create_and_add_parameter(
        self,
        group_name: str,
        station: Station,
        paths: Sequence[str],
        setter: Optional[MutableMapping[str, Any]] = None,
        getter: Optional[Callable[..., Any]] = None,
        formatter: Optional[Callable[..., Any]] = None,
        unit: Optional[str] = None,
        **kwargs: Any
    ) -> None:
        """Create delegate parameter that links to a given set of paths
        (e.g. my_instrument.my_param) on the station"""
        source_parameters = [
            self.parse_instrument_path(station, path) for path in paths
        ]
        parameter_names = self._parameter_names(source_parameters)

        setter_fn = None
        if setter is not None:
            setter_method = self.parse_instrument_path(
                station, setter.pop("method")
            )
            setter_fn = partial(setter_method, **setter)

        params = [
            self._add_parameter(group_name, name, source)
            for name, source in zip(parameter_names, source_parameters)
        ]

        group = DelegateGroup(
            name=group_name,
            parameters=params,
            parameter_names=parameter_names,
            setter=setter_fn,
            getter=getter,
            formatter=formatter
        )

        self.add_parameter(
            name=group_name,
            parameter_class=GroupedParameter,
            group=group,
            unit=unit,
            **kwargs
        )

    def _create_and_add_channels(
        self,
        station: Station,
        channels: Mapping[str, Union[str, Mapping[str, Any]]],
    ) -> None:
        """Add channels to the instrument."""
        channel_wrapper = None
        chnnls_dict: Dict[str, Union[str, Mapping[str, Any]]] = dict(channels)
        channel_type_global = chnnls_dict.pop("type", None)
        if channel_type_global is not None and \
           not isinstance(channel_type_global, str):
            raise ValueError("Wrong channel type.")
        channel_wrapper_global = _get_channel_wrapper_class(
            channel_type_global
        )

        for channel_name, input_params in chnnls_dict.items():
            if isinstance(input_params, Mapping):
                input_params = dict(input_params)
                channel_type_individual = input_params.pop("type", None)
                channel_wrapper_individual = _get_channel_wrapper_class(
                    channel_type_individual
                )
                if channel_wrapper_individual is None:
                    channel_wrapper = channel_wrapper_global
                else:
                    channel_wrapper = channel_wrapper_individual
            else:
                channel_wrapper = channel_wrapper_global

            self._create_and_add_channel(
                channel_name=channel_name,
                station=station,
                input_params=input_params,
                channel_wrapper=channel_wrapper,
            )

    def _create_and_add_channel(
        self,
        channel_name: str,
        station: Station,
        input_params: Union[str, Mapping[str, Any]],
        channel_wrapper: Optional[Type[InstrumentChannel]],
        **kwargs: Any,
    ) -> None:
        """Adds a channel to the instrument."""
        if isinstance(input_params, str):
            try:
                channel = self.parse_instrument_path(station, input_params)
            except ValueError as v_err:
                msg = "Unknown channel path."
                raise ValueError(msg) from v_err

        elif isinstance(input_params, Mapping) and channel_wrapper is not None:
            channel = self.parse_instrument_path(
                station, input_params["channel"]
            )
            wrapper_kwargs = dict(**kwargs, **input_params)

            channel = channel_wrapper(
                parent=channel.parent,
                name=channel_name,
                **wrapper_kwargs
            )
        else:
            raise ValueError(
                "Channels can only be created from existing channels, "
                "or using a wrapper channel class; "
                f"instead got {input_params!r} inputs with "
                f"{channel_wrapper!r} channel wrapper."
            )

        self.add_submodule(channel_name, channel)

    def __repr__(self) -> str:
        params = ", ".join(self.parameters.keys())
        return f"DelegateInstrument(name={self.name}, parameters={params})"


def _get_channel_wrapper_class(
    channel_type: Optional[str],
) -> Optional[Type[InstrumentChannel]]:
    """Get channel class from string specified in yaml."""
    if channel_type is None:
        return None
    else:
        channel_type_elems = str(channel_type).split(".")
        module_name = ".".join(channel_type_elems[:-1])
        instr_class_name = channel_type_elems[-1]
        module = importlib.import_module(module_name)
        return getattr(module, instr_class_name)



"""A Sphinx extension that uses Parso to extract the code of a instance attribute.

The code is used to produce a proxy object with a repr containing the code defining the
attribute. This enables better documentation of instance attributes.
Especially QCoDeS Parameters. Note that this is for the moment limited to
attributes on QCoDeS instruments."""

import functools
import inspect
from typing import Any, Dict, Optional, Tuple, Type, Union

import parso
from sphinx.util import logging
from sphinx.util.inspect import safe_getattr

from qcodes.instrument.base import InstrumentBase

LOGGER = logging.getLogger(__name__)


class ParameterProxy:

    """
    An object that acts as a proxy for documenting containing
    a repr that can be set from a string.

    """

    def __init__(self, repr_str: str):
        self._repr = repr_str

    def __repr__(self) -> str:
        """
        A repr based on the string set in init.
        """
        return self._repr


def find_class(
    node: parso.tree.BaseNode, classname: str
) -> Tuple[parso.python.tree.Class, ...]:
    """Find all classes in a given Parso node named ``classname``."""
    nodes = []
    for child in node.children:
        if isinstance(child, parso.python.tree.Class) and child.name.value == classname:
            nodes.append(child)
        elif isinstance(child, parso.python.tree.Node):
            nodes.extend(find_class(child, classname))
    return tuple(nodes)


def find_init_func(
    node: parso.tree.BaseNode,
) -> Tuple[parso.python.tree.Function, ...]:
    """Find all ``__init__`` functions in the supplied Parso node."""
    nodes = []
    for child in node.children:
        if (
            isinstance(child, parso.python.tree.Function)
            and child.name.value == "__init__"
        ):
            nodes.append(child)
        elif isinstance(child, parso.python.tree.Node):
            nodes.extend(find_init_func(child))
    return tuple(nodes)


def parse_init_function_from_str(
    code: str, classname: str
) -> Optional[parso.python.tree.Function]:
    module = parso.parse(code)
    classes = find_class(module, classname)
    if len(classes) > 1:

        LOGGER.warning(
            f"Found more than one class definition for {classname}: Found {classes}"
        )
        return None
    if len(classes) == 0:
        LOGGER.debug(f"Could not find a class definition for {classname}")
        return None
    init_funcs = find_init_func(classes[0])
    if len(init_funcs) > 1:
        LOGGER.warning(
            f"Found more than one init function for {classname}: " f"Found {init_funcs}"
        )
        return None
    if len(init_funcs) == 0:
        LOGGER.debug(f"Found no init function for {classname}")
        return None
    return init_funcs[0]


def extract_statements_from_node(
    parso_node: parso.tree.BaseNode,
) -> Tuple[parso.python.tree.ExprStmt, ...]:
    nodes = []
    for child in parso_node.children:
        if isinstance(child, parso.python.tree.ExprStmt):
            nodes.append(child)
        elif isinstance(child, parso.python.tree.Node):
            nodes.extend(extract_statements_from_node(child))
    return tuple(nodes)


@functools.lru_cache(maxsize=None, typed=True)
def eval_params_from_code(code: str, classname: str) -> Dict[str, ParameterProxy]:
    init_func_tree = parse_init_function_from_str(code, classname)
    if init_func_tree is None:
        return {}
    stms = extract_statements_from_node(init_func_tree)
    param_dict = {}

    for stm in stms:
        try:
            name_code = extract_code_as_repr(stm)
        except Exception:
            LOGGER.warning(f"Error while trying to parse attribute from {classname}")
            continue
        if name_code is not None:
            name, proxy_param = name_code
            param_dict[name] = proxy_param
    return param_dict


def extract_code_as_repr(
    stm: parso.python.tree.ExprStmt,
) -> Optional[Tuple[str, ParameterProxy]]:
    lhs = stm.children[0]
    rhs = stm.get_rhs()

    if isinstance(lhs, parso.python.tree.BaseNode) and len(lhs.children) == 2:
        obj1 = lhs.children[0]
        obj2 = lhs.children[1]
        if (
            isinstance(obj1, parso.python.tree.Leaf)
            and obj1.value == "self"
            and isinstance(obj2, parso.python.tree.BaseNode)
            and isinstance(obj2.children[1], parso.python.tree.Leaf)
        ):
            name = obj2.children[1].value
            code = " ".join(rhs.get_code().strip().split())
            pp = ParameterProxy(code)
            return name, pp
        else:
            return None
    else:
        return None


def qcodes_parameter_attr_getter(
    object_to_document_attr_on: Type[object], name: str, *default: Any
) -> Any:
    """
    Try to extract an attribute as a proxy object with a repr containing the code
    if the class the attribute is bound to is a subclass of ``InstrumentBase``
    and the attribute is not private.

    Args:
        object_to_document_attr_on: The type (not instance of the object to detect the
            attribute on.
        name: Name of the attribute to look for.
        *default: Default obejct to use as a replacement if the attribute could not be
            found.

    Returns:
        Attribute looked up, proxy object containing the code of the attribute as a
        repr or a default object.
    """
    if (
        inspect.isclass(object_to_document_attr_on)
        and issubclass(object_to_document_attr_on, InstrumentBase)
        and not name.startswith("_")
    ):
        try:
            attr = safe_getattr(object_to_document_attr_on, name)
        except AttributeError:
            LOGGER.debug(
                f"Attempting to load attribute {name} on "
                f"{object_to_document_attr_on} via parsing"
            )
            mro = inspect.getmro(object_to_document_attr_on)
            attr = None
            for classobj in mro:
                try:
                    param_dict = eval_params_from_code(
                        inspect.getsource(classobj), classobj.__name__
                    )
                    if param_dict.get(name) is not None:
                        attr = param_dict[name]
                        break
                except TypeError:
                    continue
            if attr is None:
                LOGGER.debug(
                    f"Falling back to default Sphinx attribute loader for {name}"
                    f" on {object_to_document_attr_on}"
                )
                attr = safe_getattr(object_to_document_attr_on, name, default)
    else:
        attr = safe_getattr(object_to_document_attr_on, name, default)
    return attr


def setup(app: Any) -> Dict[str, Union[str, bool]]:
    """Called by sphinx to setup the extension."""
    app.setup_extension("sphinx.ext.autodoc")  # Require autodoc extension

    app.add_autodoc_attrgetter(object, qcodes_parameter_attr_getter)

    return {
        "version": "0.1",
        "parallel_read_safe": True,  # Not tested, should not be an issue
        "parallel_write_safe": True,  # Not tested, should not be an issue
    }



import gc
import os
import shutil
import tempfile
from contextlib import contextmanager
from typing import Iterator

import numpy as np
import pytest

from qcodes.dataset.descriptions.dependencies import InterDependencies_
from qcodes.dataset.descriptions.param_spec import ParamSpec, ParamSpecBase
from qcodes.dataset.measurements import Measurement
from qcodes.dataset.sqlite.database import connect
from qcodes.instrument.parameter import (
    ArrayParameter,
    Parameter,
    ParameterWithSetpoints,
)
from qcodes.tests.instrument_mocks import (
    ArraySetPointParam,
    DummyChannelInstrument,
    DummyInstrument,
    Multi2DSetPointParam,
    Multi2DSetPointParam2Sizes,
    setpoint_generator,
)
from qcodes.utils.validators import Arrays, ComplexNumbers, Numbers


@pytest.fixture(scope='function')
def empty_temp_db_connection(tmp_path):
    """
    Yield connection to an empty temporary DB file.
    """
    path = str(tmp_path / 'source.db')
    conn = connect(path)
    try:
        yield conn
    finally:
        conn.close()
        # there is a very real chance that the tests will leave open
        # connections to the database. These will have gone out of scope at
        # this stage but a gc collection may not have run. The gc
        # collection ensures that all connections belonging to now out of
        # scope objects will be closed
        gc.collect()


@pytest.fixture(scope='function')
def two_empty_temp_db_connections(tmp_path):
    """
    Yield connections to two empty files. Meant for use with the
    test_database_extract_runs
    """

    source_path = str(tmp_path / 'source.db')
    target_path = str(tmp_path / 'target.db')
    source_conn = connect(source_path)
    target_conn = connect(target_path)
    try:
        yield (source_conn, target_conn)
    finally:
        source_conn.close()
        target_conn.close()
        # there is a very real chance that the tests will leave open
        # connections to the database. These will have gone out of scope at
        # this stage but a gc collection may not have run. The gc
        # collection ensures that all connections belonging to now out of
        # scope objects will be closed
        gc.collect()


@contextmanager
def temporarily_copied_DB(filepath: str, **kwargs):
    """
    Make a temporary copy of a db-file and delete it after use. Meant to be
    used together with the old version database fixtures, lest we change the
    fixtures on disk. Yields the connection object

    Args:
        filepath: path to the db-file

    Kwargs:
        kwargs to be passed to connect
    """
    with tempfile.TemporaryDirectory() as tmpdir:
        dbname_new = os.path.join(tmpdir, 'temp.db')
        shutil.copy2(filepath, dbname_new)

        conn = connect(dbname_new, **kwargs)

        try:
            yield conn

        finally:
            conn.close()


@pytest.fixture
def scalar_dataset(dataset):
    n_params = 3
    n_rows = 10**3
    params_indep = [ParamSpecBase(f'param_{i}',
                                  'numeric',
                                  label=f'param_{i}',
                                  unit='V')
                    for i in range(n_params)]
    param_dep = ParamSpecBase(f'param_{n_params}',
                              'numeric',
                              label=f'param_{n_params}',
                              unit='Ohm')

    all_params = params_indep + [param_dep]

    idps = InterDependencies_(dependencies={param_dep: tuple(params_indep)})

    dataset.set_interdependencies(idps)
    dataset.mark_started()
    dataset.add_results([{p.name: int(n_rows*10*pn+i)
                          for pn, p in enumerate(all_params)}
                         for i in range(n_rows)])
    dataset.mark_completed()
    yield dataset


@pytest.fixture
def scalar_dataset_with_nulls(dataset):
    """
    A very simple dataset. A scalar is varied, and two parameters are measured
    one by one
    """
    sp = ParamSpecBase('setpoint', 'numeric')
    val1 = ParamSpecBase('first_value', 'numeric')
    val2 = ParamSpecBase('second_value', 'numeric')

    idps = InterDependencies_(dependencies={val1: (sp,), val2: (sp,)})
    dataset.set_interdependencies(idps)

    dataset.mark_started()

    dataset.add_results([{sp.name: 0, val1.name: 1},
                         {sp.name: 0, val2.name: 2}])
    dataset.mark_completed()
    yield dataset


@pytest.fixture(scope="function",
                params=["array", "numeric"])
def array_dataset(experiment, request):
    meas = Measurement()
    param = ArraySetPointParam()
    meas.register_parameter(param, paramtype=request.param)

    with meas.run() as datasaver:
        datasaver.add_result((param, param.get(),))
    try:
        yield datasaver.dataset
    finally:
        datasaver.dataset.conn.close()


@pytest.fixture(scope="function",
                params=["array", "numeric"])
def array_dataset_with_nulls(experiment, request):
    """
    A dataset where two arrays are measured, one as a function
    of two other (setpoint) arrays, the other as a function of just one
    of them
    """
    meas = Measurement()
    meas.register_custom_parameter('sp1', paramtype=request.param)
    meas.register_custom_parameter('sp2', paramtype=request.param)
    meas.register_custom_parameter('val1', paramtype=request.param,
                                   setpoints=('sp1', 'sp2'))
    meas.register_custom_parameter('val2', paramtype=request.param,
                                   setpoints=('sp1',))

    with meas.run() as datasaver:
        sp1_vals = np.arange(0, 5)
        sp2_vals = np.arange(5, 10)
        val1_vals = np.ones(5)
        val2_vals = np.zeros(5)
        datasaver.add_result(('sp1', sp1_vals),
                             ('sp2', sp2_vals),
                             ('val1', val1_vals))
        datasaver.add_result(('sp1', sp1_vals),
                             ('val2', val2_vals))
    try:
        yield datasaver.dataset
    finally:
        datasaver.dataset.conn.close()


@pytest.fixture(scope="function",
                params=["array", "numeric"])
def multi_dataset(experiment, request):
    meas = Measurement()
    param = Multi2DSetPointParam()

    meas.register_parameter(param, paramtype=request.param)

    with meas.run() as datasaver:
        datasaver.add_result((param, param.get(),))
    try:
        yield datasaver.dataset
    finally:
        datasaver.dataset.conn.close()


@pytest.fixture(scope="function",
                params=["array"])
def different_setpoint_dataset(experiment, request):
    meas = Measurement()
    param = Multi2DSetPointParam2Sizes()

    meas.register_parameter(param, paramtype=request.param)

    with meas.run() as datasaver:
        datasaver.add_result((param, param.get(),))
    try:
        yield datasaver.dataset
    finally:
        datasaver.dataset.conn.close()


@pytest.fixture(scope="function")
def array_in_scalar_dataset(experiment):
    meas = Measurement()
    scalar_param = Parameter('scalarparam', set_cmd=None)
    param = ArraySetPointParam()
    meas.register_parameter(scalar_param)
    meas.register_parameter(param, setpoints=(scalar_param,),
                            paramtype='array')

    with meas.run() as datasaver:
        for i in range(1, 10):
            scalar_param.set(i)
            datasaver.add_result((scalar_param, scalar_param.get()),
                                 (param, param.get()))
    try:
        yield datasaver.dataset
    finally:
        datasaver.dataset.conn.close()


@pytest.fixture(scope="function")
def varlen_array_in_scalar_dataset(experiment):
    meas = Measurement()
    scalar_param = Parameter('scalarparam', set_cmd=None)
    param = ArraySetPointParam()
    meas.register_parameter(scalar_param)
    meas.register_parameter(param, setpoints=(scalar_param,),
                            paramtype='array')
    np.random.seed(0)
    with meas.run() as datasaver:
        for i in range(1, 10):
            scalar_param.set(i)
            param.setpoints = (np.arange(i),)
            datasaver.add_result((scalar_param, scalar_param.get()),
                                 (param, np.random.rand(i)))
    try:
        yield datasaver.dataset
    finally:
        datasaver.dataset.conn.close()


@pytest.fixture(scope="function")
def array_in_scalar_dataset_unrolled(experiment):
    """
    This fixture yields a dataset where an array-valued parameter is registered
    as a 'numeric' type and has an additional single-valued setpoint. We
    expect data to be saved as individual scalars, with the scalar setpoint
    repeated.
    """
    meas = Measurement()
    scalar_param = Parameter('scalarparam', set_cmd=None)
    param = ArraySetPointParam()
    meas.register_parameter(scalar_param)
    meas.register_parameter(param, setpoints=(scalar_param,),
                            paramtype='numeric')

    with meas.run() as datasaver:
        for i in range(1, 10):
            scalar_param.set(i)
            datasaver.add_result((scalar_param, scalar_param.get()),
                                 (param, param.get()))
    try:
        yield datasaver.dataset
    finally:
        datasaver.dataset.conn.close()


@pytest.fixture(scope="function",
                params=["array", "numeric"])
def array_in_str_dataset(experiment, request):
    meas = Measurement()
    scalar_param = Parameter('textparam', set_cmd=None)
    param = ArraySetPointParam()
    meas.register_parameter(scalar_param, paramtype='text')
    meas.register_parameter(param, setpoints=(scalar_param,),
                            paramtype=request.param)

    with meas.run() as datasaver:
        for i in ['A', 'B', 'C']:
            scalar_param.set(i)
            datasaver.add_result((scalar_param, scalar_param.get()),
                                 (param, param.get()))
    try:
        yield datasaver.dataset
    finally:
        datasaver.dataset.conn.close()


@pytest.fixture
def some_paramspecbases():

    psb1 = ParamSpecBase('psb1', paramtype='text', label='blah', unit='')
    psb2 = ParamSpecBase('psb2', paramtype='array', label='', unit='V')
    psb3 = ParamSpecBase('psb3', paramtype='array', label='', unit='V')
    psb4 = ParamSpecBase('psb4', paramtype='numeric', label='number', unit='')

    return (psb1, psb2, psb3, psb4)


@pytest.fixture
def some_paramspecs():
    """
    Some different paramspecs for testing. The idea is that we just add a
    new group of paramspecs as the need arises
    """

    groups = {}

    # A valid group. Corresponding to a heatmap with a text label at each point
    first = {}
    first['ps1'] = ParamSpec('ps1', paramtype='numeric', label='Raw Data 1',
                             unit='V')
    first['ps2'] = ParamSpec('ps2', paramtype='array', label='Raw Data 2',
                             unit='V')
    first['ps3'] = ParamSpec('ps3', paramtype='text', label='Axis 1',
                             unit='', inferred_from=[first['ps1']])
    first['ps4'] = ParamSpec('ps4', paramtype='numeric', label='Axis 2',
                             unit='V', inferred_from=[first['ps2']])
    first['ps5'] = ParamSpec('ps5', paramtype='numeric', label='Signal',
                             unit='Conductance',
                             depends_on=[first['ps3'], first['ps4']])
    first['ps6'] = ParamSpec('ps6', paramtype='text', label='Goodness',
                             unit='', depends_on=[first['ps3'], first['ps4']])
    groups[1] = first

    # a small, valid group
    second = {}
    second['ps1'] = ParamSpec('ps1', paramtype='numeric',
                              label='setpoint', unit='Hz')
    second['ps2'] = ParamSpec('ps2', paramtype='numeric', label='signal',
                              unit='V', depends_on=[second['ps1']])
    groups[2] = second

    return groups


@pytest.fixture
def some_interdeps():
    """
    Some different InterDependencies_ objects for testing
    """
    idps_list = []
    ps1 = ParamSpecBase('ps1', paramtype='numeric', label='Raw Data 1',
                        unit='V')
    ps2 = ParamSpecBase('ps2', paramtype='array', label='Raw Data 2',
                        unit='V')
    ps3 = ParamSpecBase('ps3', paramtype='text', label='Axis 1',
                        unit='')
    ps4 = ParamSpecBase('ps4', paramtype='numeric', label='Axis 2',
                        unit='V')
    ps5 = ParamSpecBase('ps5', paramtype='numeric', label='Signal',
                        unit='Conductance')
    ps6 = ParamSpecBase('ps6', paramtype='text', label='Goodness',
                        unit='')

    idps = InterDependencies_(dependencies={ps5: (ps3, ps4), ps6: (ps3, ps4)},
                              inferences={ps4: (ps2,), ps3: (ps1,)})

    idps_list.append(idps)

    ps1 = ParamSpecBase('ps1', paramtype='numeric',
                        label='setpoint', unit='Hz')
    ps2 = ParamSpecBase('ps2', paramtype='numeric', label='signal',
                        unit='V')
    idps = InterDependencies_(dependencies={ps2: (ps1,)})

    idps_list.append(idps)

    return idps_list


@pytest.fixture  # scope is "function" per default
def DAC():
    dac = DummyInstrument('dummy_dac', gates=['ch1', 'ch2'])
    yield dac
    dac.close()


@pytest.fixture  # scope is "function" per default
def DAC_with_metadata():
    dac = DummyInstrument('dummy_dac', gates=['ch1', 'ch2'],
                          metadata={"dac": "metadata"})
    yield dac
    dac.close()


@pytest.fixture
def DMM():
    dmm = DummyInstrument('dummy_dmm', gates=['v1', 'v2'])
    yield dmm
    dmm.close()


@pytest.fixture
def channel_array_instrument():
    channelarrayinstrument = DummyChannelInstrument('dummy_channel_inst')
    yield channelarrayinstrument
    channelarrayinstrument.close()


@pytest.fixture
def complex_num_instrument():

    class MyParam(Parameter):

        def get_raw(self):
            return self.instrument.setpoint() + 1j*self.instrument.setpoint()

    class RealPartParam(Parameter):

        def get_raw(self):
            return self.instrument.complex_setpoint().real

    dummyinst = DummyInstrument('dummy_channel_inst', gates=())

    dummyinst.add_parameter('setpoint',
                            parameter_class=Parameter,
                            initial_value=0,
                            label='Some Setpoint',
                            unit="Some Unit",
                            vals=Numbers(),
                            get_cmd=None, set_cmd=None)

    dummyinst.add_parameter('complex_num',
                            parameter_class=MyParam,
                            initial_value=0+0j,
                            label='Complex Num',
                            unit="complex unit",
                            vals=ComplexNumbers(),
                            get_cmd=None, set_cmd=None)

    dummyinst.add_parameter('complex_setpoint',
                            initial_value=0+0j,
                            label='Complex Setpoint',
                            unit="complex unit",
                            vals=ComplexNumbers(),
                            get_cmd=None, set_cmd=None)

    dummyinst.add_parameter('real_part',
                            parameter_class=RealPartParam,
                            label='Real Part',
                            unit="real unit",
                            vals=Numbers(),
                            set_cmd=None)

    dummyinst.add_parameter('some_array_setpoints',
                            label='Some Array Setpoints',
                            unit='some other unit',
                            vals=Arrays(shape=(5,)),
                            set_cmd=False,
                            get_cmd=lambda: np.arange(5))

    dummyinst.add_parameter('some_array',
                            parameter_class=ParameterWithSetpoints,
                            setpoints=(dummyinst.some_array_setpoints,),
                            label='Some Array',
                            unit='some_array_unit',
                            vals=Arrays(shape=(5,)),
                            get_cmd=lambda: np.ones(5),
                            set_cmd=False)

    dummyinst.add_parameter('some_complex_array_setpoints',
                            label='Some complex array setpoints',
                            unit='some_array_unit',
                            get_cmd=lambda: np.arange(5),
                            set_cmd=False)

    dummyinst.add_parameter('some_complex_array',
                            label='Some Array',
                            unit='some_array_unit',
                            get_cmd=lambda: np.ones(5) + 1j*np.ones(5),
                            set_cmd=False)

    yield dummyinst
    dummyinst.close()


@pytest.fixture
def SpectrumAnalyzer():
    """
    Yields a DummyInstrument that holds ArrayParameters returning
    different types
    """

    class Spectrum(ArrayParameter):

        def __init__(self, name, instrument, **kwargs):
            super().__init__(
                name=name,
                shape=(1,),  # this attribute should be removed
                label="Flower Power Spectrum",
                unit="V/sqrt(Hz)",
                setpoint_names=("Frequency",),
                setpoint_units=("Hz",),
                instrument=instrument,
                **kwargs,
            )

            self.npts = 100
            self.start = 0
            self.stop = 2e6

        def get_raw(self):
            # This is how it should be: the setpoints are generated at the
            # time we get. But that will of course not work with the old Loop
            self.setpoints = (tuple(np.linspace(self.start, self.stop,
                                                self.npts)),)
            # not the best SA on the market; it just returns noise...
            return np.random.randn(self.npts)

    class MultiDimSpectrum(ArrayParameter):

        def __init__(self, name, instrument, **kwargs):
            self.start = 0
            self.stop = 2e6
            self.npts = (100, 50, 20)
            sp1 = np.linspace(self.start, self.stop,
                              self.npts[0])
            sp2 = np.linspace(self.start, self.stop,
                              self.npts[1])
            sp3 = np.linspace(self.start, self.stop,
                              self.npts[2])
            setpoints = setpoint_generator(sp1, sp2, sp3)

            super().__init__(
                name=name,
                instrument=instrument,
                setpoints=setpoints,
                shape=(100, 50, 20),
                label="Flower Power Spectrum in 3D",
                unit="V/sqrt(Hz)",
                setpoint_names=("Frequency0", "Frequency1", "Frequency2"),
                setpoint_units=("Hz", "Other Hz", "Third Hz"),
                **kwargs,
            )

        def get_raw(self):
            return np.random.randn(*self.npts)

    class ListSpectrum(Spectrum):

        def get_raw(self):
            output = super().get_raw()
            return list(output)

    class TupleSpectrum(Spectrum):

        def get_raw(self):
            output = super().get_raw()
            return tuple(output)

    SA = DummyInstrument('dummy_SA')
    SA.add_parameter('spectrum', parameter_class=Spectrum)
    SA.add_parameter('listspectrum', parameter_class=ListSpectrum)
    SA.add_parameter('tuplespectrum', parameter_class=TupleSpectrum)
    SA.add_parameter('multidimspectrum', parameter_class=MultiDimSpectrum)
    yield SA

    SA.close()


@pytest.fixture
def meas_with_registered_param(experiment, DAC, DMM):
    meas = Measurement()
    meas.register_parameter(DAC.ch1)
    meas.register_parameter(DMM.v1, setpoints=[DAC.ch1])
    yield meas


@pytest.fixture(name='dummyinstrument')
def _make_dummy_instrument() -> Iterator[DummyChannelInstrument]:
    inst = DummyChannelInstrument('dummyinstrument')
    try:
        yield inst
    finally:
        inst.close()


class ArrayshapedParam(Parameter):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_raw(self):
        shape = self.vals.shape

        return np.random.rand(*shape)



"""
Test properties of the coordinate transforms in the field vector module to
test if everything has been correctly implemented.
"""
import numpy as np
import json
from hypothesis import given, settings
from hypothesis.strategies import floats
from hypothesis.strategies import tuples

from qcodes.math_utils.field_vector import FieldVector
from qcodes.utils.helpers import NumpyJSONEncoder

random_coordinates = {
    "cartesian": tuples(
        floats(min_value=0, max_value=1),  # x
        floats(min_value=0, max_value=1),  # y
        floats(min_value=0, max_value=1)   # z
    ),
    "spherical": tuples(
        floats(min_value=0, max_value=1),    # r
        floats(min_value=0, max_value=180),  # theta
        floats(min_value=0, max_value=180)   # phi
    ),
    "cylindrical": tuples(
        floats(min_value=0, max_value=1),    # rho
        floats(min_value=0, max_value=180),  # phi
        floats(min_value=0, max_value=1)     # z
    )
}


@given(random_coordinates["spherical"])
@settings(max_examples=10)
def test_spherical_properties(spherical0):
    """
    If the cartesian to spherical transform has been correctly implemented
    then we expect certain symmetrical properties
    """
    # Generate a random coordinate in spherical representation
    spherical0 = dict(zip(["r", "theta", "phi"], spherical0))
    cartisian0 = FieldVector(**spherical0).get_components("x", "y", "z")

    # Mirror the theta angle in the xy-plane.
    # This should flip the sign of the z-coordinate
    spherical1 = dict(spherical0)
    spherical1["theta"] = 180 - spherical0["theta"]
    cartisian1 = FieldVector(**spherical1).get_components("x", "y", "z")
    assert np.allclose(cartisian0, cartisian1 * np.array([1, 1, -1]))

    # Add 180 to the phi coordinate.
    # This should flip the sign of the xy coordinate
    spherical2 = dict(spherical0)
    spherical2["phi"] = 180 + spherical0["phi"]
    cartisian2 = FieldVector(**spherical2).get_components("x", "y", "z")
    assert np.allclose(cartisian0, cartisian2 * np.array([-1, -1, 1]))

    # Mirroring the theta angle in the xy-plane
    # and adding 180 to the phi coordinate
    # should flip all cartesian coordinates
    spherical3 = dict(spherical0)
    spherical3["theta"] = 180 - spherical0["theta"]  # should only flip z
    spherical3["phi"] = 180 + spherical0["phi"]  # should flip xy
    cartisian3 = FieldVector(**spherical3).get_components("x", "y", "z")
    assert np.allclose(cartisian0, cartisian3 * np.array([-1, -1, -1]))

    # Finally, flipping the sign of the r coordinate
    # should flip all cartesian coordinates
    spherical4 = dict(spherical0)
    spherical4["r"] = -spherical0["r"]  # This should flip all coordinates
    cartisian4 = FieldVector(**spherical4).get_components("x", "y", "z")
    assert np.allclose(cartisian0, cartisian4 * np.array([-1, -1, -1]))


@given(random_coordinates["cylindrical"])
@settings(max_examples=10)
def test_cylindrical_properties(cylindrical0):
    """
    If the cartesian to cylindrical transform has been correctly implemented
    then we expect certain symmetrical properties
    """
    # Generate a random coordinate in cylindrical representation
    cylindrical0 = dict(zip(["rho", "phi", "z"], cylindrical0))
    cartisian0 = FieldVector(**cylindrical0).get_components("x", "y", "z")

    # If we flip the sign of the rho coordinate, we will flip the xy coordinate
    cylindrical1 = dict(cylindrical0)
    cylindrical1["rho"] *= -1
    cartisian1 = FieldVector(**cylindrical1).get_components("x", "y", "z")

    assert np.allclose(cartisian0, cartisian1 * np.array([-1, -1, 1]))


@given(
    random_coordinates["cylindrical"],
    random_coordinates["spherical"]
)
@settings(max_examples=10)
def test_triangle_inequality(cylindrical0, spherical0):
    cylindrical0 = FieldVector(**dict(zip(["rho", "phi", "z"], cylindrical0)))
    spherical0 = FieldVector(**dict(zip(["r", "phi", "theta"], spherical0)))

    assert (cylindrical0 + spherical0).norm() \
           <= (cylindrical0.norm() + spherical0.norm())
    assert cylindrical0.distance(spherical0) \
           <= (cylindrical0.norm() + spherical0.norm())


@given(random_coordinates["cartesian"])
@settings(max_examples=10)
def test_homogeneous_roundtrip(cartesian0):
    vec = FieldVector(**dict(zip("xyz", cartesian0)))
    h_vec = 13 * vec.as_homogeneous()

    assert np.allclose(
        vec.get_components(*"xyz"),
        FieldVector.from_homogeneous(h_vec).get_components(*"xyz")
    )


@given(random_coordinates["spherical"])
@settings(max_examples=10)
def test_json_dump(spherical0):
    vec = FieldVector(**dict(zip(["r", "phi", "theta"], spherical0)))
    dump = json.dumps(vec, cls=NumpyJSONEncoder)

    assert json.loads(dump) == {
        '__class__': FieldVector.__name__,
        '__args__': [vec.x, vec.y, vec.z]
    }


def test_all_attributes_are_floats():
    cartesian0 = (400, 200, 300)
    cylindrical0 = (1, 52, 0)
    spherical0 = (1, 78, 145)

    cartesian = FieldVector(**dict(zip("xyz", cartesian0)))
    cylindrical = FieldVector(**dict(zip(["rho", "phi", "z"], cylindrical0)))
    spherical = FieldVector(**dict(zip(["r", "phi", "theta"], spherical0)))

    # Test that all attributes are floats upon creation
    for fv in [cartesian, cylindrical, spherical]:
        for attr in FieldVector.attributes:
            assert isinstance(getattr(fv, attr), float)

    # Test that all attributes are floats even after setting components
    for fv in [cartesian, cylindrical, spherical]:
        for set_comp in FieldVector.attributes:
            fv.set_component(**{set_comp: 1})

            for attr in FieldVector.attributes:
                assert isinstance(getattr(fv, attr), float)



import pytest
from qcodes.utils.validators import Ints, Lists


def test_type():
    l = Lists()
    v1 = ['a', 'b', 5]
    l.validate(v1)

    v2 = 234
    with pytest.raises(TypeError):
        l.validate(v2)


def test_elt_vals():
    l = Lists(Ints(max_value=10))
    v1 = [0, 1, 5]
    l.validate(v1)

    v2 = [0, 1, 11]
    with pytest.raises(ValueError):
        l.validate(v2)


def test_valid_values():
    val = Lists(Ints(max_value=10))
    for vval in val.valid_values:
        val.validate(vval)



import numpy as np
from typing import Any, Tuple, Dict, Union

from qcodes import (
    VisaInstrument, InstrumentChannel, Parameter, ParameterWithSetpoints
)
from qcodes.instrument.parameter import ParamRawDataType
from qcodes.utils.validators import Enum, Numbers, Arrays, Ints
from qcodes.utils.helpers import create_on_off_val_mapping


class FrequencyAxis(Parameter):

    def __init__(self,
                 start: Parameter,
                 stop: Parameter,
                 npts: Parameter,
                 *args: Any,
                 **kwargs: Any
                 ) -> None:
        super().__init__(*args, **kwargs)
        self._start: Parameter = start
        self._stop: Parameter = stop
        self._npts: Parameter = npts

    def get_raw(self) -> ParamRawDataType:
        start_val = self._start()
        stop_val = self._stop()
        npts_val = self._npts()
        assert start_val is not None
        assert stop_val is not None
        assert npts_val is not None
        return np.linspace(start_val, stop_val, npts_val)


class Trace(ParameterWithSetpoints):

    def __init__(self, number: int, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.instrument: Union["SpectrumAnalyzerMode", "PhaseNoiseMode"]
        self.root_instrument: "N9030B"

        self.number = number

    def get_raw(self) -> ParamRawDataType:
        return self.instrument._get_data(trace_num=self.number)


class SpectrumAnalyzerMode(InstrumentChannel):
    """
    Spectrum Analyzer Mode for Keysight N9030B instrument.
    """

    def __init__(self, parent: "N9030B", name: str, *arg: Any, **kwargs: Any):
        super().__init__(parent, name, *arg, **kwargs)

        self._min_freq = -8e7
        self._valid_max_freq: Dict[str, float] = {"503": 3.7e9,
                                                  "508": 8.5e9,
                                                  "513": 13.8e9,
                                                  "526": 27e9,
                                                  "544": 44.5e9}
        opt: str
        for hw_opt_for_max_freq in self._valid_max_freq.keys():
            if hw_opt_for_max_freq in self.root_instrument._options():
                opt = hw_opt_for_max_freq
        self._max_freq = self._valid_max_freq[opt]

        self.add_parameter(
            name="start",
            unit="Hz",
            get_cmd=":SENSe:FREQuency:STARt?",
            set_cmd=self._set_start,
            get_parser=float,
            vals=Numbers(self._min_freq, self._max_freq - 10),
            docstring="start frequency for the sweep"
        )

        self.add_parameter(
            name="stop",
            unit="Hz",
            get_cmd=":SENSe:FREQuency:STOP?",
            set_cmd=self._set_stop,
            get_parser=float,
            vals=Numbers(self._min_freq + 10, self._max_freq),
            docstring="stop frequency for the sweep"
        )

        self.add_parameter(
            name="center",
            unit="Hz",
            get_cmd=":SENSe:FREQuency:CENTer?",
            set_cmd=self._set_center,
            get_parser=float,
            vals=Numbers(self._min_freq + 5, self._max_freq - 5),
            docstring="Sets and gets center frequency"
        )

        self.add_parameter(
            name="span",
            unit="Hz",
            get_cmd=":SENSe:FREQuency:SPAN?",
            set_cmd=self._set_span,
            get_parser=float,
            vals=Numbers(10, self._max_freq - self._min_freq),
            docstring="Changes span of frequency"
        )

        self.add_parameter(
            name="npts",
            get_cmd=":SENSe:SWEep:POINts?",
            set_cmd=self._set_npts,
            get_parser=int,
            vals=Ints(1, 20001),
            docstring="Number of points for the sweep"
        )

        self.add_parameter(
            name="sweep_time",
            label="Sweep time",
            get_cmd=":SENSe:SWEep:TIME?",
            set_cmd=":SENSe:SWEep:TIME {}",
            get_parser=float,
            unit="s",
            docstring="gets sweep time"
        )

        self.add_parameter(
            name="auto_sweep_time_enabled",
            get_cmd=":SENSe:SWEep:TIME:AUTO?",
            set_cmd=self._enable_auto_sweep_time,
            val_mapping=create_on_off_val_mapping(on_val="ON", off_val="OFF"),
            docstring="enables auto sweep time"
        )

        self.add_parameter(
            name="auto_sweep_type_enabled",
            get_cmd=":SENSe:SWEep:TYPE:AUTO?",
            set_cmd=self._enable_auto_sweep_type,
            val_mapping=create_on_off_val_mapping(on_val="ON", off_val="OFF"),
            docstring="enables auto sweep type"
        )

        self.add_parameter(
            name="sweep_type",
            get_cmd=":SENSe:SWEep:TYPE?",
            set_cmd=self._set_sweep_type,
            val_mapping={
                "fft": "FFT",
                "sweep": "SWE",
            },
            docstring="Sets up sweep type. Possible options are 'fft' and "
                      "'sweep'."
        )

        self.add_parameter(
            name="freq_axis",
            label="Frequency",
            unit="Hz",
            start=self.start,
            stop=self.stop,
            npts=self.npts,
            vals=Arrays(shape=(self.npts.get_latest,)),
            parameter_class=FrequencyAxis,
            docstring="Creates frequency axis for the sweep from start, "
                      "stop and npts values."
        )

        self.add_parameter(
            name="trace",
            label="Trace",
            unit="dB",
            number=1,
            vals=Arrays(shape=(self.npts.get_latest,)),
            setpoints=(self.freq_axis,),
            parameter_class=Trace,
            docstring="Gets trace data."
        )

    def _set_start(self, val: float) -> None:
        """
        Sets start frequency
        """
        stop = self.stop()
        if val >= stop:
            raise ValueError(f"Start frequency must be smaller than stop "
                             f"frequency. Provided start freq is: {val} Hz and "
                             f"set stop freq is: {stop} Hz")

        self.write(f":SENSe:FREQuency:STARt {val}")

        start = self.start()
        if abs(val - start) >= 1:
            self.log.warning(
                f"Could not set start to {val} setting it to {start}"
            )

    def _set_stop(self, val: float) -> None:
        """
        Sets stop frequency
        """
        start = self.start()
        if val <= start:
            raise ValueError(f"Stop frequency must be larger than start "
                             f"frequency. Provided stop freq is: {val} Hz and "
                             f"set start freq is: {start} Hz")

        self.write(f":SENSe:FREQuency:STOP {val}")

        stop = self.stop()
        if abs(val - stop) >= 1:
            self.log.warning(
                f"Could not set stop to {val} setting it to {stop}"
            )

    def _set_center(self, val: float) -> None:
        """
        Sets center frequency and updates start and stop frequencies if they
        change.
        """
        self.write(f":SENSe:FREQuency:CENTer {val}")
        self.update_trace()

    def _set_span(self, val: float) -> None:
        """
        Sets frequency span and updates start and stop frequencies if they
        change.
        """
        self.write(f":SENSe:FREQuency:SPAN {val}")
        self.update_trace()

    def _set_npts(self, val: int) -> None:
        """
        Sets number of points for sweep
        """
        self.write(f":SENSe:SWEep:POINts {val}")

    def _enable_auto_sweep_time(self, val: str) -> None:
        """
        Enables auto sweep time
        """
        self.write(f":SENSe:SWEep:TIME:AUTO {val}")

    def _enable_auto_sweep_type(self, val: str) -> None:
        """
        Enables auto sweep type
        """
        self.write(f":SENSe:SWEep:TYPE:AUTO {val}")

    def _set_sweep_type(self, val: str) -> None:
        """
        Sets sweep type
        """
        self.write(f":SENSe:SWEep:TYPE {val}")

    def _get_data(self, trace_num: int) -> ParamRawDataType:
        """
        Gets data from the measurement.
        """
        try:
            timeout = self.sweep_time() + self.root_instrument._additional_wait
            with self.root_instrument.timeout.set_to(timeout):
                data_str = self.ask(f":READ:"
                                    f"{self.root_instrument.measurement()}"
                                    f"{trace_num}?")
                data = np.array(data_str.rstrip().split(",")).astype("float64")
        except TimeoutError as e:
            raise TimeoutError("Couldn't receive any data. Command timed "
                               "out.") from e

        trace_data = data[1::2]
        return trace_data

    def update_trace(self) -> None:
        """
        Updates start and stop frequencies whenever span of/or center frequency
        is updated.
        """
        self.start()
        self.stop()

    def setup_swept_sa_sweep(self,
                             start: float,
                             stop: float,
                             npts: int) -> None:
        """
        Sets up the Swept SA measurement sweep for Spectrum Analyzer Mode.
        """
        self.root_instrument.mode("SA")
        if "SAN" in self.root_instrument._available_meas():
            self.root_instrument.measurement("SAN")
        else:
            raise RuntimeError("Swept SA measurement is not available on your "
                               "Keysight N9030B instrument with Spectrum "
                               "Analyzer mode.")
        self.start(start)
        self.stop(stop)
        self.npts(npts)

    def autotune(self) -> None:
        """
        Autotune quickly get to the most likely signal of interest, and
        position it optimally on the display.
        """
        self.write(":SENS:FREQuency:TUNE:IMMediate")
        self.center()


class PhaseNoiseMode(InstrumentChannel):
    """
    Phase Noise Mode for Keysight N9030B instrument.
    """

    def __init__(self, parent: "N9030B", name: str, *arg: Any, **kwargs: Any):
        super().__init__(parent, name, *arg, **kwargs)

        self._min_freq = 1
        self._valid_max_freq: Dict[str, float] = {"503": 3699999995,
                                                  "508": 8499999995,
                                                  "513": 13799999995,
                                                  "526": 26999999995,
                                                  "544": 44499999995}
        opt: str
        for hw_opt_for_max_freq in self._valid_max_freq.keys():
            if hw_opt_for_max_freq in self.root_instrument._options():
                opt = hw_opt_for_max_freq
        self._max_freq = self._valid_max_freq[opt]

        self.add_parameter(
            name="npts",
            get_cmd=":SENSe:LPLot:SWEep:POINts?",
            set_cmd=":SENSe:LPLot:SWEep:POINts {}",
            get_parser=int,
            vals=Ints(601, 20001),
            docstring="Number of points for the sweep"
        )

        self.add_parameter(
            name="start_offset",
            unit="Hz",
            get_cmd=":SENSe:LPLot:FREQuency:OFFSet:STARt?",
            set_cmd=self._set_start_offset,
            get_parser=float,
            vals=Numbers(self._min_freq, self._max_freq - 10),
            docstring="start frequency offset for the plot"
        )

        self.add_parameter(
            name="stop_offset",
            unit="Hz",
            get_cmd=":SENSe:LPLot:FREQuency:OFFSet:STOP?",
            set_cmd=self._set_stop_offset,
            get_parser=float,
            vals=Numbers(self._min_freq + 99, self._max_freq),
            docstring="stop frequency offset for the plot"
        )

        self.add_parameter(
            name="signal_tracking_enabled",
            get_cmd=":SENSe:FREQuency:CARRier:TRACk?",
            set_cmd=":SENSe:FREQuency:CARRier:TRACk {}",
            val_mapping=create_on_off_val_mapping(on_val="ON", off_val="OFF"),
            docstring="Gets/Sets signal tracking. When signal tracking is "
                      "enabled carrier signal is repeatedly realigned. Signal "
                      "Tracking assumes the new acquisition occurs repeatedly "
                      "without pause."
        )

        self.add_parameter(
            name="freq_axis",
            label="Frequency",
            unit="Hz",
            start=self.start_offset,
            stop=self.stop_offset,
            npts=self.npts,
            vals=Arrays(shape=(self.npts.get_latest,)),
            parameter_class=FrequencyAxis,
            docstring="Creates frequency axis for the sweep from "
                      "start_offset, stop_offset and npts values."
        )

        self.add_parameter(
            name="trace",
            label="Trace",
            unit="dB",
            number=3,
            vals=Arrays(shape=(self.npts.get_latest,)),
            setpoints=(self.freq_axis,),
            parameter_class=Trace,
            docstring="Gets trace data."
        )

    def _set_start_offset(self, val: float) -> None:
        """
        Sets start offset for frequency in the plot
        """
        stop_offset = self.stop_offset()
        self.write(f":SENSe:LPLot:FREQuency:OFFSet:STARt {val}")
        start_offset = self.start_offset()

        if abs(val - start_offset) >= 1:
            self.log.warning(
                f"Could not set start offset to {val} setting it to "
                f"{start_offset}"
            )
        if val >= stop_offset or abs(val - stop_offset) < 10:
            self.log.warning(f"Provided start frequency offset {val} Hz was "
                             f"greater than preset stop frequency offset "
                             f"{stop_offset} Hz. Provided start frequency "
                             f"offset {val} Hz is set and new stop freq offset"
                             f" is: {self.stop_offset()} Hz.")

    def _set_stop_offset(self, val: float) -> None:
        """
        Sets stop offset for frequency in the plot
        """
        start_offset = self.start_offset()
        self.write(f":SENSe:LPLot:FREQuency:OFFSet:STOP {val}")
        stop_offset = self.stop_offset()

        if abs(val - stop_offset) >= 1:
            self.log.warning(
                f"Could not set stop offset to {val} setting it to "
                f"{stop_offset}"
            )

        if val <= start_offset or abs(val-start_offset) < 10:
            self.log.warning(f"Provided stop frequency offset {val} Hz was "
                             f"less than preset start frequency offset "
                             f"{start_offset} Hz. Provided stop frequency "
                             f"offset {val} Hz is set and new start freq offset"
                             f" is: {self.start_offset()} Hz.")

    def _get_data(self, trace_num: int) -> ParamRawDataType:
        """
        Gets data from the measurement.
        """
        raw_data = self.ask(f":READ:{self.root_instrument.measurement()}{1}?")
        trace_res_details = np.array(
            raw_data.rstrip().split(",")
        ).astype("float64")

        if len(trace_res_details) != 7 or (
                len(trace_res_details) >= 1 and trace_res_details[0] < -50
        ):
            self.log.warning("Carrier(s) Incorrect or Missing!")
            return -1 * np.ones(self.npts())

        try:
            data_str = self.ask(f":READ:{self.root_instrument.measurement()}"
                                f"{trace_num}?")
            data = np.array(data_str.rstrip().split(",")).astype("float64")
        except TimeoutError as e:
            raise TimeoutError("Couldn't receive any data. Command timed "
                               "out.") from e

        trace_data = data[1::2]
        return trace_data

    def setup_log_plot_sweep(self,
                             start_offset: float,
                             stop_offset: float,
                             npts: int
                             ) -> None:
        """
        Sets up the Log Plot measurement sweep for Phase Noise Mode.
        """
        self.root_instrument.mode("PNOISE")
        if "LPL" in self.root_instrument._available_meas():
            self.root_instrument.measurement("LPL")
        else:
            raise RuntimeError("Log Plot measurement is not available on your "
                               "Keysight N9030B instrument with Phase Noise "
                               "mode.")

        self.start_offset(start_offset)
        self.stop_offset(stop_offset)
        self.npts(npts)

    def autotune(self) -> None:
        """
        On autotune, the measurement automatically searches for and tunes to
        the strongest signal in the full span of the analyzer.
        """
        self.write(":SENSe:FREQuency:CARRier:SEARch")
        self.start_offset()
        self.stop_offset()


class N9030B(VisaInstrument):
    """
    Driver for Keysight N9030B PXA signal analyzer. Keysight N9030B PXA
    siganl analyzer is part of Keysight X-Series Multi-touch Signal
    Analyzers.
    This driver allows Swept SA measurements in Spectrum Analyzer mode and
    Log Plot measurements in Phase Noise mode of the instrument.

    Args:
        name
        address
    """

    def __init__(self, name: str, address: str, **kwargs: Any) -> None:
        super().__init__(name, address, terminator='\n', **kwargs)

        self._min_freq: float
        self._max_freq: float
        self._additional_wait: float = 1

        self.add_parameter(
            name="mode",
            get_cmd=":INSTrument:SELect?",
            set_cmd=":INSTrument:SELect {}",
            vals=Enum(*self._available_modes()),
            docstring="Allows setting of different modes present and licensed "
                      "for the instrument."
        )

        self.add_parameter(
            name="measurement",
            get_cmd=":CONFigure?",
            set_cmd=":CONFigure:{}",
            vals=Enum("SAN", "LPL"),
            docstring="Sets measurement type from among the available "
                      "measurement types."
        )

        self.add_parameter(
            name="cont_meas",
            initial_value=False,
            get_cmd=":INITiate:CONTinuous?",
            set_cmd=self._enable_cont_meas,
            val_mapping=create_on_off_val_mapping(on_val="ON", off_val="OFF"),
            docstring="Enables or disables continuous measurement."
        )

        self.add_parameter(
            name="format",
            get_cmd=":FORMat:TRACe:DATA?",
            set_cmd=":FORMat:TRACe:DATA {}",
            val_mapping={
                "ascii": "ASCii",
                "int32": "INTeger,32",
                "real32": "REAL,32",
                "real64": "REAL,64"
            },
            docstring="Sets up format of data received"
        )

        if "SA" in self._available_modes():
            sa_mode = SpectrumAnalyzerMode(self, name="sa")
            self.add_submodule("sa", sa_mode)
        else:
            self.log.info("Spectrum Analyzer mode is not available on this "
                          "instrument.")

        if "PNOISE" in self._available_modes():
            pnoise_mode = PhaseNoiseMode(self, name="pn")
            self.add_submodule("pn", pnoise_mode)
        else:
            self.log.info("Phase Noise mode is not available on this "
                          "instrument.")
        self.connect_message()

    def _available_modes(self) -> Tuple[str, ...]:
        """
        Returns present and licensed modes for the instrument.
        """
        available_modes = self.ask(":INSTrument:CATalog?")
        av_modes = available_modes[1:-1].split(',')
        modes: Tuple[str, ...] = ()
        for i, mode in enumerate(av_modes):
            if i == 0:
                modes = modes + (mode.split(' ')[0], )
            else:
                modes = modes + (mode.split(' ')[1], )
        return modes

    def _available_meas(self) -> Tuple[str, ...]:
        """
        Gives available measurement with a given mode for the instrument
        """
        available_meas = self.ask(":CONFigure:CATalog?")
        av_meas = available_meas[1:-1].split(',')
        measurements: Tuple[str, ...] = ()
        for i, meas in enumerate(av_meas):
            if i == 0:
                measurements = measurements + (meas, )
            else:
                measurements = measurements + (meas[1:], )
        return measurements

    def _enable_cont_meas(self, val: str) -> None:
        """
        Sets continuous measurement to ON or OFF.
        """
        self.write(f":INITiate:CONTinuous {val}")

    def _options(self) -> Tuple[str, ...]:
        """
        Returns installed options numbers.
        """
        options_raw = self.ask('*OPT?')
        return tuple(options_raw[1:-1].split(','))

    def reset(self) -> None:
        """
        Reset the instrument by sending the RST command
        """
        self.write("*RST")

    def abort(self) -> None:
        """
        Aborts the measurement
        """
        self.write(":ABORt")



import numpy

from qcodes.data.data_array import DataArray
from qcodes.data.data_set import new_data
from qcodes.data.io import DiskIO


class MockFormatter:
    def read(self, data_set):
        data_set.has_read_data = True

    def write(self, data_set, io_manager, location, write_metadata=False):
        data_set.has_written_data = True

    def read_metadata(self, data_set):
        data_set.has_read_metadata = True

    def write_metadata(self, data_set, io_manager, location, read_first=True):
        data_set.has_written_metadata = True


class RecordingMockFormatter:
    def __init__(self):
        self.write_calls = []
        self.modified_ranges = []
        self.last_saved_indices = []
        self.write_metadata_calls = []

    def write(self, data_set, io_manager, location, force_write=False):
        self.write_calls.append((io_manager.base_location, location))

        self.modified_ranges.append({
            array_id: array.modified_range
            for array_id, array in data_set.arrays.items()
        })

        self.last_saved_indices.append({
            array_id: array.last_saved_index
            for array_id, array in data_set.arrays.items()
        })

    def write_metadata(self, data_set, io_manager, location, read_first=True):
        self.write_metadata_calls.append((io_manager.base_location,
                                          location, read_first))


class MatchIO:
    def __init__(self, existing_matches, fmt=None):
        self.existing_matches = existing_matches
        self.fmt = fmt or '{}{}.something'

    def list(self, location, **kwargs):
        return [self.fmt.format(location, i) for i in self.existing_matches]

    def join(self, *args):
        return DiskIO('.').join(*args)


class MockLive:
    arrays = 'whole lotta data'


class MockArray:
    array_id = 'noise'

    def init_data(self):
        self.ready = True


def DataSet1D(location=None, name=None):
    # DataSet with one 1D array with 5 points

    # TODO: since y lists x as a set_array, it should automatically
    # set is_setpoint=True for x, shouldn't it? Any reason we woundn't
    # want that?
    x = DataArray(name='x', label='X', preset_data=(1., 2., 3., 4., 5.),
                  is_setpoint=True)
    y = DataArray(name='y', label='Y', preset_data=(3., 4., 5., 6., 7.),
                  set_arrays=(x,))
    return new_data(arrays=(x, y), location=location, name=name)


def DataSet2D(location=None, name=None):
    # DataSet with one 2D array with 4 x 6 points
    yy, xx = numpy.meshgrid(range(4), range(6))
    zz = xx**2+yy**2
    # outer setpoint should be 1D
    xx = xx[:, 0]
    x = DataArray(name='x', label='X', preset_data=xx, is_setpoint=True)
    y = DataArray(name='y', label='Y', preset_data=yy, set_arrays=(x,),
                  is_setpoint=True)
    z = DataArray(name='z', label='Z', preset_data=zz, set_arrays=(x, y))
    return new_data(arrays=(x, y, z), location=location, name=name)


def file_1d():
    return '\n'.join([
        '# x_set\ty',
        '# "X"\t"Y"',
        '# 5',
        '1\t3',
        '2\t4',
        '3\t5',
        '4\t6',
        '5\t7', ''])


def DataSetCombined(location=None, name=None):
    # Complex DataSet with two 1D and two 2D arrays
    x = DataArray(name='x', label='X!', preset_data=(16., 17.),
                  is_setpoint=True)
    y1 = DataArray(name='y1', label='Y1 value', preset_data=(18., 19.),
                   set_arrays=(x,))
    y2 = DataArray(name='y2', label='Y2 value', preset_data=(20., 21.),
                   set_arrays=(x,))

    yset = DataArray(name='y', label='Y', preset_data=(22., 23., 24.),
                     is_setpoint=True)
    yset.nest(2, 0, x)
    z1 = DataArray(name='z1', label='Z1',
                   preset_data=((25., 26., 27.), (28., 29., 30.)),
                   set_arrays=(x, yset))
    z2 = DataArray(name='z2', label='Z2',
                   preset_data=((31., 32., 33.), (34., 35., 36.)),
                   set_arrays=(x, yset))
    return new_data(arrays=(x, y1, y2, yset, z1, z2), location=location,
                    name=name)


def files_combined():
    return [
        '\n'.join([
            '# x_set\ty1\ty2',
            '# "X!"\t"Y1 value"\t"Y2 value"',
            '# 2',
            '16\t18\t20',
            '17\t19\t21', '']),

        '\n'.join([
            '# x_set\ty_set\tz1\tz2',
            '# "X!"\t"Y"\t"Z1"\t"Z2"',
            '# 2\t3',
            '16\t22\t25\t31',
            '16\t23\t26\t32',
            '16\t24\t27\t33',
            '',
            '17\t22\t28\t34',
            '17\t23\t29\t35',
            '17\t24\t30\t36', ''])
    ]



# Driver for microwave source HP_83650A
#
# Written by Bruno Buijtendorp (brunobuijtendorp@gmail.com)
from typing import Optional, Any

import logging
from qcodes import VisaInstrument
from qcodes import validators as vals

log = logging.getLogger(__name__)


def parsestr(v: str) -> str:
    return v.strip().strip('"')


class HP_83650A(VisaInstrument):

    def __init__(self,
                 name: str,
                 address: str,
                 verbose: int = 1,
                 reset: bool = False,
                 server_name: Optional[str] = None,
                 **kwargs: Any):
        """ Driver for HP_83650A

        """
        self.verbose = verbose
        log.debug('Initializing instrument')
        super().__init__(name, address, **kwargs)

        self.add_parameter('frequency',
                           label='Frequency',
                           get_cmd='FREQ:CW?',
                           set_cmd='FREQ:CW {}',
                           vals=vals.Numbers(10e6, 40e9),
                           docstring='Microwave frequency, ....',
                           get_parser=float,
                           unit='Hz')

        self.add_parameter('freqmode',
                           label='Frequency mode',
                           get_cmd='FREQ:MODE?',
                           set_cmd='FREQ:MODE {}',
                           vals=vals.Strings(),
                           get_parser=parsestr,
                           docstring='Microwave frequency mode, ....')

        self.add_parameter('power',
                           label='Power',
                           get_cmd='SOUR:POW?',
                           set_cmd='SOUR:POW {}',
                           vals=vals.Numbers(-20, 20),
                           get_parser=float,
                           unit='dBm',
                           docstring='Microwave power, ....')

        self.add_parameter('rfstatus',
                           label='RF status',
                           get_cmd=':POW:STAT?',
                           set_cmd=':POW:STAT {}',
                           val_mapping={'on': '1', 'off': '0'},
                           vals=vals.Strings(),
                           get_parser=parsestr,
                           docstring='Status, ....')

        self.add_parameter('fmstatus',
                           label='FM status',
                           get_cmd=':FM:STAT?',
                           set_cmd=':FM:STAT {}',
                           val_mapping={'on': '1', 'off': '0'},
                           vals=vals.Strings(),
                           get_parser=parsestr,
                           docstring='FM status, ....')

        self.add_parameter('fmcoup',
                           label='FM coupling',
                           get_cmd=':FM:COUP?',
                           set_cmd=':FM:COUP {}',
                           vals=vals.Strings(),
                           get_parser=parsestr,
                           docstring='FM coupling, ....')

        self.add_parameter('amstatus',
                           label='AM status',
                           get_cmd=':AM:STAT?',
                           set_cmd=':AM:STAT {}',
                           val_mapping={'on': '1', 'off': '0'},
                           vals=vals.Strings(),
                           get_parser=parsestr,
                           docstring='AM status, ....')

        self.add_parameter('pulsestatus',
                           label='Pulse status',
                           get_cmd=':PULS:STAT?',
                           set_cmd=':PULS:STAT {}',
                           val_mapping={'on': '1', 'off': '0'},
                           vals=vals.Strings(),
                           get_parser=parsestr,
                           docstring='Pulse status, ....')

        self.add_parameter('pulsesource',
                           label='Pulse source',
                           get_cmd=':PULS:SOUR?',
                           set_cmd=':PULS:SOUR {}',
                           vals=vals.Strings(),
                           get_parser=parsestr,
                           docstring='Pulse source, ....')
        self.connect_message()

    def reset(self) -> None:
        log.debug('Resetting instrument')
        self.write('*RST')
        self.print_all()

    def print_all(self) -> None:
        log.debug('Reading all settings from instrument')
        print(self.rfstatus.label + ':', self.rfstatus.get())
        print(self.power.label + ':', self.power.get(), self.power.unit)
        print(self.frequency.label +
              ': %e' % self.frequency.get(), self.frequency.unit)
        print(self.freqmode.label + ':', self.freqmode.get())
        self.print_modstatus()

    def print_modstatus(self) -> None:
        print(self.fmstatus.label + ':', self.fmstatus.get())
        print(self.fmcoup.label + ':', self.fmcoup.get())
        print(self.amstatus.label + ':', self.amstatus.get())
        print(self.pulsestatus.label + ':', self.pulsestatus.get())
        print(self.pulsesource.label + ':', self.pulsesource.get())



import pytest
import qcodes.instrument_drivers.Keysight.Keysight_N6705B as N6705B

import qcodes.instrument.sims as sims

visalib = sims.__file__.replace('__init__.py', 'Keysight_N6705B.yaml@sim')


@pytest.fixture(scope='module')
def driver():
    driver = N6705B.N6705B('N6705B',
                           address="GPIB::1::INSTR",
                           visalib=visalib)
    yield driver
    driver.close()


def test_idn(driver):
    assert {'firmware': 'D.01.08',
            'model': 'N6705B',
            'serial': 'MY50001897',
            'vendor': 'Agilent Technologies'} == driver.IDN()


def test_channels(driver):
    # Ensure each channel got instantiated
    assert len(driver.channels) == 4
    for (i, ch) in enumerate(driver.channels):
        assert(ch.channel == i+1)
        assert(ch.ch_name == f"ch{i+1}")



import logging
from .hdf5_format import HDF5Format
import hickle
from qcodes.utils.helpers import deep_update

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .data_set import DataSet


#%%

log = logging.getLogger(__name__)


class HDF5FormatHickle(HDF5Format):

    _metadata_file = 'snapshot.hickle'
    _format_tag = 'hdf5-hickle'

    def write_metadata(self, data_set: 'DataSet', io_manager=None, location=None, read_first=False, **kwargs):
        """
        Write all metadata in this DataSet to storage.

        Args:
            data_set: the data we're storing

            io_manager (io_manager): the base location to write to

            location (str): the file location within io_manager

            read_first (Optional[bool]): read previously saved metadata before
                writing? The current metadata will still be the used if
                there are changes, but if the saved metadata has information
                not present in the current metadata, it will be retained.
                Default True.
        """

        # this statement is here to make the linter happy
        if io_manager is None or location is None:
            raise Exception('please set io_manager and location arguments ')

        if read_first:
            # In case the saved file has more metadata than we have here,
            # read it in first. But any changes to the in-memory copy should
            # override the saved file data.
            memory_metadata = data_set.metadata
            data_set.metadata = {}
            self.read_metadata(data_set)
            deep_update(data_set.metadata, memory_metadata)

        log.info('writing metadata to file %s' % self._metadata_file)
        fn = io_manager.join(location, self._metadata_file)
        with io_manager.open(fn, 'w', encoding='utf8') as snap_file:
            hickle.dump(data_set.metadata, snap_file)

    def read_metadata(self, data_set: 'DataSet'):
        """ Reads in the metadata

        Args:
            data_set: Dataset object to read the metadata into
        """
        io_manager = data_set.io
        location = data_set.location
        fn = io_manager.join(location, self._metadata_file)
        if io_manager.list(fn):
            log.info('reading metadata from file %s' % self._metadata_file)
            with io_manager.open(fn, 'r') as snap_file:
                metadata = hickle.load(snap_file)
            data_set.metadata.update(metadata)



from qcodes import VisaInstrument, InstrumentChannel
from typing import Union, List, Tuple, Optional


class KeysightSubModule(InstrumentChannel):
    """
    A base class for submodules for the 34980A systems.

    Args:
        parent: the system which the module is installed on
        name: user defined name for the module
        slot: the slot the module is installed
    """
    def __init__(
            self,
            parent: Union[VisaInstrument, InstrumentChannel],
            name: str,
            slot: int
    ) -> None:

        super().__init__(parent, name)

        self.slot = slot


class KeysightSwitchMatrixSubModule(KeysightSubModule):
    """
    A base class for **Switch Matrix** submodules for the 34980A systems.
    """
    def validate_value(self, row: int, column: int) -> None:
        """
        to check if the row and column number is within the range of the module
        layout.

        Args:
            row: row value
            column: column value
        """
        raise NotImplementedError("Please subclass this")

    def to_channel_list(
            self,
            paths: List[Tuple[int, int]],
            wiring_config: Optional[str] = None
    ) -> str:
        """
        convert the (row, column) pair to a 4-digit channel number 'sxxx', where
        s is the slot number, xxx is generated from the numbering function.
        This may be different for different modules.

        Args:
            paths: list of channels to connect [(r1, c1), (r2, c2), (r3, c3)]
            wiring_config: for 1-wire matrices, values are 'MH', 'ML';
                                 for 2-wire matrices, values are 'M1H', 'M2H',
                                 'M1L', 'M2L'

        Returns:
            in the format of '(@sxxx, sxxx, sxxx, sxxx)', where sxxx is a
            4-digit channel number
        """
        raise NotImplementedError("Please subclass this")

    def is_open(self, row: int, column: int) -> bool:
        """
        to check if a channel is open/disconnected

        Args:
            row: row number
            column: column number

        Returns:
            True if the channel is open/disconnected
            False if it's closed/connected.
        """
        self.validate_value(row, column)
        channel = self.to_channel_list([(row, column)])
        message = self.ask(f'ROUT:OPEN? {channel}')
        return bool(int(message))

    def is_closed(self, row: int, column: int) -> bool:
        """
        to check if a channel is closed/connected

        Args:
            row: row number
            column: column number

        Returns:
            True if the channel is closed/connected
            False if it's open/disconnected.
        """
        self.validate_value(row, column)
        channel = self.to_channel_list([(row, column)])
        message = self.ask(f'ROUT:CLOSe? {channel}')
        return bool(int(message))

    def connect(self, row: int, column: int) -> None:
        """
        to connect/close the specified channels

        Args:
            row: row number
            column: column number
        """
        self.validate_value(row, column)
        channel = self.to_channel_list([(row, column)])
        self.write(f'ROUT:CLOSe {channel}')

    def disconnect(self, row: int, column: int) -> None:
        """
        to disconnect/open the specified channels

        Args:
            row: row number
            column: column number
        """
        self.validate_value(row, column)
        channel = self.to_channel_list([(row, column)])
        self.write(f'ROUT:OPEN {channel}')

    def connect_paths(self, paths: List[Tuple[int, int]]) -> None:
        """
        to connect/close the specified channels.

        Args:
            paths: list of channels to connect [(r1, c1), (r2, c2), (r3, c3)]
        """
        for row, column in paths:
            self.validate_value(row, column)
        channel_list_str = self.to_channel_list(paths)
        self.write(f"ROUTe:CLOSe {channel_list_str}")

    def disconnect_paths(self, paths: List[Tuple[int, int]]) -> None:
        """
        to disconnect/open the specified channels.

        Args:
            paths: list of channels to connect [(r1, c1), (r2, c2), (r3, c3)]
        """
        for row, column in paths:
            self.validate_value(row, column)
        channel_list_str = self.to_channel_list(paths)
        self.write(f"ROUTe:OPEN {channel_list_str}")

    def are_closed(self, paths: List[Tuple[int, int]]) -> List[bool]:
        """
        to check if a list of channels is closed/connected

        Args:
            paths: list of channels [(r1, c1), (r2, c2), (r3, c3)]

        Returns:
            a list of True and/or False
            True if the channel is closed/connected
            False if it's open/disconnected.
        """
        for row, column in paths:
            self.validate_value(row, column)
        channel_list_str = self.to_channel_list(paths)
        messages = self.ask(f"ROUTe:CLOSe? {channel_list_str}")
        return [bool(int(message)) for message in messages.split(',')]

    def are_open(self, paths: List[Tuple[int, int]]) -> List[bool]:
        """
        to check if a list of channels is open/disconnected

        Args:
            paths: list of channels [(r1, c1), (r2, c2), (r3, c3)]

        Returns:
            a list of True and/or False
            True if the channel is closed/connected
            False if it's open/disconnected.
        """
        for row, column in paths:
            self.validate_value(row, column)
        channel_list_str = self.to_channel_list(paths)
        messages = self.ask(f"ROUTe:OPEN? {channel_list_str}")
        return [bool(int(message)) for message in messages.split(',')]
